#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Defines a base class for AIXI-approximate agents.
"""

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import copy
import random

import util

# Define a enumeration to represent what type of environment update has been performed.
update_enum = util.enum('action_update', 'percept_update')

# Define some short cuts for ease of reference.
action_update = update_enum.action_update
percept_update = update_enum.percept_update

class Agent:
    """ This base class represents the minimum class elements for a AIXI-style agent.

        The following attributes and methods must be available in all agents, in order
        for the main interaction loop to get responses, give environmental feedback,
        manage learning, and monitor progress:

         - `age`
         - `average_reward()`
         - `generate_random_action()`
         - `last_update`
         - `learning_period`
         - `maximum_action()`
         - `maximum_reward()`
         - `model_size()`
         - `model_update_action()`
         - `model_update_percept()`
         - `search()`
         - `total_reward`
    """

    # Instance methods.

    def __init__(self, environment = None, options = {}):
        """ Construct an AIXI-style learning agent from the given configuration values and the environment.

             - `environment` is an instance of the pyaixi.Environment class that the agent with interact with.
             - `options` is a dictionary of named options and their values.

            The following options are optional:
             - `learning-period`: the number of cycles the agent should learn for.
                                  Defaults to '0', which is indefinite learning.
        """

        # The number of interaction cycles the agent has been alive.
        # Set initially to 0.
        # (Also called `time_cycle` in the C++ version.)
        self.age = 0

        # A reference to the environment the agent interacts with.
        # Set to the environment given. Mandatory.
        # (Called `m_environment` in the C++ version.)
        assert environment is not None, "A non-null environment is required."
        self.environment = environment

        # The type of the last update (action or percept).
        # Set initial to 'action_update'.
        # (Called `m_last_update` in the C++ version.)
        self.last_update = action_update

        # The number of cycles during which the agent learns.
        # Retrieved from the given options under 'learning-period'. Defaults to 0 if not given.
        # (Called `m_learning_period` in the C++ version.)
        self.learning_period = int(options.get('learning-period', 0))

        # Stores the given configuration options.
        # (Called `m_options` in the C++ version.)
        self.options = options

        # The total reward earnt by this agent so far.
        # Set initially to 0.
        # (Called `totalReward` in the C++ version.)
        self.total_reward = 0
    # end def

    def average_reward(self):
        """ Returns the average reward received by the agent at each time step.
            (Called `averageReward` in the C++ version.)
        """

        # The average reward is the total reward, divided by the number of cycles.
        # (Ensure a safe default if the average can't be calculated yet.)
        if self.age > 0:
            average = self.total_reward / self.age
            return average
        else:
            return 0.0
        # end if
    # end def

    def generate_random_action(self):
        """ Returns an action generated uniformly at random.
            (Called `genRandomAction` in the C++ version.)
        """

        return random.choice(self.environment.valid_actions)
    # end def

    def maximum_action(self):
        """ Returns the maximum action the agent can execute.
            (Called `maxAction` in the C++ version.)
        """

        # Get the value from the environment.
        if self.environment is not None:
            return self.environment.maximum_action()
        else:
            return None
        # end if
    # end def

    def maximum_reward(self):
        """ Returns the maximum possible reward the agent can receive in a single cycle.

            (Called `maxReward` in the C++ version.)
        """

        # Get the value from the environment.
        if self.environment is not None:
            return self.environment.maximum_reward()
        else:
            return None
        # end if
    # end def

    def model_size(self):
        """ Returns the size of the agent's model.

            WARNING: this method should be overriden by inheriting classes.

            (Called `modelSize` in the C++ version.)
        """
        return 0
    # end def

    def model_update_action(self, action):
        """ Update the agent's model of the world with an action from the
            environment.

            - `action`: the action that was performed.

            WARNING: this method should be overriden by inheriting classes.

            (Called `modelUpdate` in the C++ version.)
        """
        pass
    # end def

    def model_update_percept(self, observation, reward):
        """ Update the agent's model of the world with a percept from the
            environment.

            - `observation`: the observation that was received.
            - `reward`: the reward that was received.

            WARNING: this method should be overriden by inheriting classes.

            (Called `modelUpdate` in the C++ version.)
        """
        pass
    # end def

    def search(self):
        """ Returns the best action for this agent.
        """

        return self.maximum_action()
    # end def

    def reset(self):
        """ Resets the agent.

            NOTE: this method may need to be overriden by inheriting classes,
                  with this method called using `Agent.reset(self)`.
        """
        # Reset the current time cycle, total rewards, and last update action appropriately.
        self.age = 0
        self.total_reward = 0.0
        self.last_update = action_update
    # end def
# end class


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Defines an environment for an extended version of an agent interacting with an environment
where there's a tiger and a pot of gold hidden separately, behind two closed doors.
"""

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import random

from pyaixi import environment, util

# Define a enumeration to represent agent interactions with the environment,
# such as listening, just standing, or opening doors.
extended_tiger_action_enum = util.enum('aListen', 'aLeft', 'aRight', 'aStand')

# Define a enumeration to represent environment observations: either not
# hearing the tiger, or hearing it from behind either door.
extended_tiger_observation_enum = util.enum('oNull', 'oLeft', 'oRight')

# Define a enumeration to represent rewards as a result of actions: being eaten by the tiger,
# performing an invalid action (listening while sitting), getting information from listening,
# just standing there, or finding the gold.
# NOTE: since the enumeration values need to be positive, these values are defined relative to
#       100.
#       So -100 points is 0, -1 points is 99, and 30 points is 130.
extended_tiger_reward_enum = util.enum(rInvalid = 0, rTiger = 0, rStand = 99, rListen = 100, rGold = 130)

# Define some shorthand notation for ease of reference.
aListen = extended_tiger_action_enum.aListen
aLeft   = extended_tiger_action_enum.aLeft
aRight  = extended_tiger_action_enum.aRight
aStand  = extended_tiger_action_enum.aStand

oNull  = extended_tiger_observation_enum.oNull
oLeft  = extended_tiger_observation_enum.oLeft
oRight = extended_tiger_observation_enum.oRight

rInvalid  = extended_tiger_reward_enum.rInvalid
rTiger = extended_tiger_reward_enum.rTiger
rStand = extended_tiger_reward_enum.rStand
rListen = extended_tiger_reward_enum.rListen
rGold   = extended_tiger_reward_enum.rGold

class ExtendedTiger(environment.Environment):
    """ The environment is a more elaborate version of Tiger.

        There are two doors and a stool. A tiger is hidden behind one door
        and a pot of gold is hidden behind the other.

        The agent begins each round sitting on the stool where it
        may either listen for the tiger (`aListen`) or stand up
        `aStand`.

        Listening for the tiger results in an observation
        which correctly describes the tiger's whereabouts with probability
        `m_listen_accuracy` and a reward of `rListen`.

        Standing up results in an uninformative observation (`oNull`) and
        a reward of `rStand`.

        Once the agent is standing, it may open either the left or right door
        (`oLeft` and `oRight`). Doing so results in an uninformative observation
        `oNull` and a reward based on what is behind the door (`rGold` or `rTiger`).

        After opening a door the agent is re-seated and the tiger and gold randomly
        re-allocated to a door (`reset()`).

        Attempting to open a door while seated, to listen while standing, or to
        stand while already standing will result in an uninformative observation
        (`Null`) and a reward of `rInvalid`.

        Domain characteristics:

        - environment: "extended_tiger"
        - maximum action: 3 (2 bits)
        - maximum observation: 2 (2 bits)
        - maximum reward: 130 (8 bits)

        Configuration options:
        - `tiger-listen-accuracy`: the probability that listening (while seated) for the
                                   tiger results in the correct observation of the tiger's
                                   whereabouts (`listen_accuracy`).

                                   Must be a floating point number between 0.0
                                   and 1.0 inclusive.

                                   Default value is `default_listen_accuracy`, 85%.
                                   (Optional.)
    """

    # Instance attributes.

    # Set the default probability for the listen accuracy.
    default_listen_accuracy = 0.85

    # Instance methods.

    def __init__(self, options = {}):
        """ Construct the Tiger environment from the given options.

             - `options` is a dictionary of named options and their values.

            The following options in `options` are optional:
             - `tiger-listen-accuracy`: the probability of correctly observing the tiger's location from listening.
                                        Default value: 85%.
        """

        # Set up the base environment.
        environment.Environment.__init__(self, options = options)

        # Define the acceptable action values.
        self.valid_actions = extended_tiger_action_enum.keys()

        # Define the acceptable observation values.
        self.valid_observations = extended_tiger_observation_enum.keys()

        # Define the acceptable reward values.
        self.valid_rewards = extended_tiger_reward_enum.keys()

        # Set the accuracy of the listen action.
        if 'tiger-listen-accuracy' not in options:
            options["tiger-listen-accuracy"] = self.default_listen_accuracy
        # end if
        self.listen_accuracy = float(options["tiger-listen-accuracy"])

        # Make sure the accuracy value is valid.
        assert 0.0 <= self.listen_accuracy and self.listen_accuracy <= 1.0

        # Set an initial percept.
        self.observation = oNull
        self.reward = 0

        # Place the tiger, and the gold.
        self.reset()
    # end def

    def perform_action(self, action):
        """ Receives the agent's action and calculates the new environment percept.
            (Called `performAction` in the C++ version.)
        """

        assert self.is_valid_action(action)

        # Save the action.
        self.action = action

        # Unless explicitly accounted for, the action is invalid, and the observation is null.
        self.observation = oNull
        self.reward = rInvalid

        if (action == aListen and self.sitting):
            # Listen while sitting down, and return the correct door with probability
            # equal to self.listen_acurracy.
            self.observation = self.tiger if random.random() < self.listen_accuracy else self.gold
            self.reward = rListen
        elif (action == aLeft and not self.sitting):
            # Open the left door while standing. Get a reward based on what was behind
            # the door. Reseat the agent and reallocate the tiger and the gold.
            self.reward = rTiger if self.tiger == oLeft else rGold
            self.reset()
        elif (action == aRight and not self.sitting):
            # Open the right door while standing. Get a reward based on what was behind
            # the door. Reseat the agent and reallocate the tiger and the gold.
            self.reward = rTiger if self.tiger == oRight else rGold
            self.reset()
        elif (action == aStand and self.sitting):
            # Stand from a seated position. Get the reward for standing.
            self.reward = rStand
            self.sitting = False
        # end if

        return (self.observation, self.reward)
    # end def

    def print(self):
        """ Returns a string indicating the status of the environment.
        """

        action_text      = {aListen: "listen",
                            aLeft: "open left door",
                            aRight: "open right door",
                            aStand: "stand up"}
        observation_text = {oNull: "null",
                            oLeft: "hear tiger at left door",
                            oRight: "hear tiger at right door"}
        reward_text      = {rTiger: "eaten",
                            rInvalid: "invalid action",
                            rStand: "stand up",
                            rListen: "listen",
                            rGold: "gold!"}
        state_text       = {False: "standing",
                            True: "sitting"}

        # Show what just happened, correcting the reward for being defined relative to 100.
        message = "action = %s, observation = %s, reward = %s (%d), agent is now %s"% \
                  (action_text[self.action],
                   observation_text[self.observation],
                   reward_text[self.reward],
                   (self.reward - 100),
                   state_text[self.sitting])

        return message
    # end def

    def reset(self):
        """ Resets the environment by randomly placing the tiger behind one door,
            the gold behind the other, and reseating the agent.
        """

        # Place the tiger randomly.
        self.tiger = oLeft if random.random() < 0.5 else oRight

        # Place the gold behind the opposite door.
        self.gold  = oRight if self.tiger == oLeft else oLeft

        # Start the agent sitting down.
        self.sitting = True;
    # end def
# end class


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Define some helper functions.
"""

from __future__ import division
from __future__ import print_function

try:
    import collections
    IterableUserDict = collections.UserDict
except:
    import UserDict
    IterableUserDict = UserDict.IterableUserDict
# end try

def bits_required(integer_value):
    """ Return the number of bits required to store the given integer.
    """
    assert type(integer_value) == int and integer_value >= 0, "The given number must be an integer greater than or equal to zero."

    # The bin built-in function converts the value to a binary string, but adds an '0b' prefix.
    # Count the length of the string, but subtract 2 for the '0b' prefix.
    return len(bin(integer_value)) - 2
# end def

def decode(symbol_list, bit_count):
    """ Decodes the value encoded on the end of a list of symbols.
        Each symbol is a bit in the binary representation of the value, with more significant
        bits at the end of the list.

        - `symbol_list` - the list of symbols to decode from.
        - `bit_count` - the number of bits from the end of the symbol list to decode.
    """
    assert bit_count > 0, "The given number of bits (%d) is invalid." % bit_count
    assert bit_count <= len(symbol_list), "The given number of bits (%d) is greater than the length of the symbol list. (%d)" % (bit_count, len(symbol_list))

    # Take the last `bit_count` number of symbols from the end of the given symbol list.
    bits = symbol_list[-bit_count:]

    # Reverse the list of bits, and make a string out of them.
    bits.reverse()
    bit_string = ''.join(map(str, bits))

    # Return the bit string as an integer via the built-in int command, telling it that the number in the string is binary/base 2.
    return int(bit_string, 2)
# end def

def encode(integer_symbol, bit_count):
    """ Returns an updated version of the given symbol list with the given symbol encoded into binary.

        - `symbol_list` - the list onto which to encode the value.
        - `integer_symbol` - the integer value to be encoded.
        - `bit_count` - the number of bits from the end of the symbol list to decode.
    """

    assert type(integer_symbol) == int and integer_symbol >= 0, "The given symbol must be an integer greater than or equal to zero."

    # Convert the symbol into a bit string.
    bit_string = bin(integer_symbol)

    # Strip off any '0b' prefix.
    if bit_string.startswith('0b'):
        bit_string = bit_string[2:]
    # end if

    # Convert the string into a list of integers.
    bits = [int(bit) for bit in list(bit_string)]

    # Check that the number of bits is not bigger than the given bit count.
    bits_length = len(bits)
    assert bit_count >= bits_length, \
           "The given number of bits %d to encode is smaller than the bits needed to encode %d." % \
               (bit_count, bits_length)

    # Calculate how many bits we need to pad the bit string with, if any, and pad with zeros.
    pad_list = [0 for i in xrange(0, bits_length - bit_count)]

    # Return the newly created bit list, with the zero padding first.
    symbol_list = pad_list + bits
    return symbol_list
# end def

def enum(*sequential, **named):
    """ Define an enumeration type helper, since the operation of this codebase depends heavily on enumeration types.

        Usage:

        new_enum = enum('value1', 'value2')

        This code is base on code by StackOverflow user Alec Thomas http://stackoverflow.com/users/7980/alec-thomas
        From: http://stackoverflow.com/questions/36932/how-can-i-represent-an-enum-in-python
        License: CC-Wiki/CC BY-SA 3.0 with attribution.
    """

    # Construct a mapping of the sequential values to the names.
    enum_dict = dict(zip([str(value) for value in sequential], range(len(sequential))), **named)

    # Reverse this, so that we've got a way to quickly look up values to names.
    reverse = dict((value, key) for key, value in enum_dict.items())

    # Set up a dictionary (with user-modifiable attributes) from the reverse mapping,
    # so that iteration over the enumeration and membership checks are possible.
    enums = IterableUserDict(reverse)

    # Add the original and reverse mappings to the dictionary.
    enums.mapping = enum_dict
    enums.reverse_mapping = reverse

    # Make each of the name values have an attribute, for convenience.
    # e.g. new_enum.value1 == 0   new_enum.value2 == 1
    for (key, value) in enum_dict.items():
        setattr(enums, str(key), int(value))
    # end for

    # Return the generated structure.
    return enums
# end def


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Defines a class for the MC-AIXI-CTW agent.
"""

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import copy
import random

from pyaixi import agent, prediction, search, util

from pyaixi.agent import update_enum, action_update, percept_update
from pyaixi.prediction import ctw_context_tree
from pyaixi.search import monte_carlo_search_tree
from pyaixi.search.monte_carlo_search_tree import nodetype_enum, chance_node, decision_node


class MC_AIXI_CTW_Undo:
    """ A class to save details from a MC-AIXI-CTW agent to restore state later.
    """

    # Instance methods.

    def __init__(self, agent):
        """ Store values from the given agent that can be used to revert that agent to a previous state.
        """

        # Copy the main attributes of the given agent into this class.
        self.age = agent.age
        self.total_reward = agent.total_reward
        self.history_size = agent.history_size()
        self.last_update = agent.last_update
    # end def
# end class


class MC_AIXI_CTW_Agent(agent.Agent):
    """ This class represents a MC-AIXI-CTW agent.

        It includes much of the high-level logic for choosing suitable actions.
        In particular, the agent maintains an internal model of the environment using
        a context tree.

        It uses this internal model to to predict the probability of future outcomes:

         - `get_predicted_action_probability()`
         - `percept_probability()`

        as well as to generate actions and precepts according to the model distribution:

         - `generate_action()`
         - `gen_percept()`
         - `generate_percept_and_update()`
         - `generate_random_action()`

        Actions are chosen via the UCT algorithm, which is orchestrated by a
        high-level search function and a playout policy:

         - `search()`
         - `playout()`
         - `horizon`
         - `mc_simulations`
         - `search_tree`

        Several functions decode/encode actions and percepts between the
        corresponding types (i.e. `action_enum`, `percept_enum`) and generic
        representation by symbol lists:

         - `decode_action()`
         - `decode_observation()`
         - `decode_percept()`
         - `decode_reward()`
         - `encode_action()`
         - `encode_percept()`

        There are various attributes which describe the agent and its
        interaction with the environment so far:

         - `age`
         - `average_reward`
         - `history_size()`
         - `horizon`
         - `last_update`
         - `maximum_action()`
         - `maximum_bits_needed()`
         - `maximum_reward()`
         - `total_reward`
    """

    # Instance methods.

    def __init__(self, environment = None, options = {}):
        """ Construct a MC-AIXI-CTW learning agent from the given configuration values and the environment.

             - `environment` is an instance of the pyaixi.Environment class that the agent with interact with.
             - `options` is a dictionary of named options and their values.

            `options` must contain the following mandatory options:
             - `agent-horizon`: the agent's planning horizon.
             - `ct-depth`: the depth of the context tree for this agent, in symbols/bits.
             - `mc-simulations`: the number of simulations to run when choosing new actions.

            The following options are optional:
             - `learning-period`: the number of cycles the agent should learn for.
                                  Defaults to '0', which is indefinite learning.
        """

        # Set up the base agent options, which handles getting and setting the learning period, amongst other basic values.
        agent.Agent.__init__(self, environment = environment, options = options)

        # The agent's context tree depth.
        # Retrieved from the given options under 'ct-depth'. Mandatory.
        # (Called `ct_depth` in the C++ version.)
        assert 'ct-depth' in options, \
               "The required 'ct-depth' context tree depth option is missing from the given options."
        self.depth = int(options['ct-depth'])

        # (CTW) Context tree representing the agent's model of the environment.
        # Created for this instance.
        # (Called `m_ct` in the C++ version.)
        self.context_tree = ctw_context_tree.CTWContextTree(self.depth)

        # The length of the agent's planning horizon.
        # Retrieved from the given options under 'agent-horizon'. Mandatory.
        # (Called `m_horizon` in the C++ version.)
        assert 'agent-horizon' in options, \
               "The required 'agent-horizon' search horizon option is missing from the given options."
        self.horizon = int(options['agent-horizon'])

        # The number of simulations to conduct when choosing new actions via the UCT algorithm.
        # Retrieved from the given options under 'mc-simulations'. Mandatory.
        # (Called `m_mc_simulations` in the C++ version.)
        assert 'mc-simulations' in options, \
               "The required 'mc-simulations' Monte Carlo simulations count option is missing from the given options."
        self.mc_simulations = int(options['mc-simulations'])

        self.reset()
    # end def

    def decode_action(self, symbol_list):
        """ Returns the action decoded from the beginning of the given list of symbols.

            - `symbol_list`: the symbol list to decode the action from.

            (Called `decodeAction` in the C++ version.)
        """

        return util.decode(symbol_list, self.environment.action_bits())
    # end def

    def decode_observation(self, symbol_list):
        """ Returns the observation decoded from the given list of symbols.

            - `symbol_list`: the symbol list to decode the observation from.

            (Called `decodeObservation` in the C++ version.)
        """

        return util.decode(symbol_list, self.environment.observation_bits())
    # end def

    def decode_reward(self, symbol_list):
        """ Returns the reward decoded from the beginning of the given list of symbols.

            - `symbol_list`: the symbol list to decode the reward from.

            (Called `decodeReward` in the C++ version.)
        """

        return util.decode(symbol_list, self.environment.reward_bits())
    # end def

    def decode_percept(self, symbol_list):
        """ Returns the percept (observation and reward) decoded from the beginning of
            the given list of symbols.

            - `symbol_list`: the symbol list to decode the percept from.

            (Called `decodePercept` in the C++ version.)
        """

        # Check if we've got exactly enough symbols.
        reward_bits = self.environment.reward_bits()
        observation_bits = self.environment.observation_bits()

        assert len(symbol_list) >= (reward_bits + observation_bits),\
               "The given symbol list isn't long enough to contain a percept."

        # Get the reward symbols from the given symbol list, starting with the
        # reward, then getting the observation from the list after that.
        reward_symbols      = symbol_list[:reward_bits]
        observation_symbols = symbol_list[reward_bits:(reward_bits + observation_bits)]

        # Decode the obtained symbols.
        reward      = self.decode_reward(reward_symbols)
        observation = self.decode_observation(observation_symbols)

        # Return the decoded percept as a tuple of observation and reward.
        return (observation, reward)
    # end def

    def encode_action(self, action):
        """ Returns the given action encoded as a list of symbols.

            - `action`: the action to encode.

            (Called `encodeAction` in the C++ version.)
        """

        return util.encode(action, self.environment.action_bits())
    # end def

    def encode_percept(self, observation, reward):
        """ Returns the given percept (an observation, reward part) as a list of symbols.

            - `observation`: the observation part of the percept to encode.
            - `reward`: the reward part of the percept to encode.
    
            (Called `encode_percept` in the C++ version.)
        """

        # Add first the encoded reward, then the encoded observation to the list of output symbols.
        symbol_list  = util.encode(reward, self.environment.reward_bits())
        symbol_list += util.encode(observation, self.environment.observation_bits())

        # Return the generated list.
        return symbol_list
    # end def

    def generate_action(self):
        """ Returns an action generated according to the agent's history
            statistics by doing rejection sampling from the context tree.

            (Called `genAction` in the C++ version.)
        """

        assert self.last_update == percept_update, \
               "The last action must have been a percept update before generate_action can be called."

        # Sample from the context tree, getting a list of symbols corresponding to an action.
        # NOTE: works only for a binary symbol alphabet.
        action_list = self.context_tree.generate_random_symbols(self.environment.action_bits())

        # Decode the sample into an action, then return it.
        return self.decode_action(action_list)
    # end def

    def generate_percept(self):
        """ Returns a percept (an observation, reward pair) distributed according to the agent's history
            statistics by sampling from the context tree.

            (Called `genPercept` in the C++ version.)
        """

        # Sample from the context tree.
        percept_symbols = self.context_tree.generate_random_symbols(self.environment.percept_bits())

        # Decode and return the percept symbols into the desired observation and reward.
        return self.decodePercept(percept_symbols)
    # end def

    def generate_percept_and_update(self):
        """ Returns a percept (an observation, reward pair) distributed according to the agent's history
            statistics, after updating the context tree with it.

            (Called `genPerceptAndUpdate` in the C++ version.)
        """

        # Sample from the context tree.
        percept_symbols = self.context_tree.generate_random_symbols_and_update(self.environment.percept_bits())

        # Decode the percept symbols into the desired observation and reward.
        observation, reward = self.decode_percept(percept_symbols)

        # Update other properties.
        self.total_reward += reward
        self.last_update = percept_update

        # Return the observation and reward as the promised tuple.
        return (observation, reward)
    # end def

    def get_predicted_action_probability(self, action):
        """ Returns the probability of selecting a particular action according to the
            agent's internal model of its own behaviour.

            - `action`: the action we wish to find the likelihood of.

            (Called `getPredictedActionProb` in the C++ version.)
        """
        # Do some sanity checks.
        assert self.environment.is_valid_action(action), "The given action is invalid."
        assert self.last_update == percept_update, "Can only make predictions after a percept update."

        # Encode the action first.
        action_symbols = self.encode_action(action)

        # Get the probability of seeing this action using the context tree 
        # to make a prediction.
        return self.context_tree.predict(action_symbols)
    # end def

    def history_size(self):
        """ Returns the length of the stored history for an agent.
            (Called `historySize` in the C++ version.)
        """

        return len(self.context_tree.history)
    # end def

    def maximum_bits_needed(self):
        """ Returns the maximum number of bits needed to represent actions or percepts.
            NOTE: this is for binary alphabets.

            (Called `maxBitsNeeded` in the C++ version.)
        """

        return max(self.environment.action_bits(), self.environment.percept_bits())
    # end def

    def model_revert(self, undo_instance):
        """ Revert the agent's internal model of the world to that of a previous time cycle,
            using the given undo class instance.

            (Called `modelRevert` in the C++ version.)
        """

        # Revert excess actions and percepts.
        while self.history_size() > undo_instance.history_size:
            # Undo an action or a percept, depending on the current state of the agent.
            if self.last_update == percept_update:
                # Undo a percept.
                self.context_tree.revert(self.environment.percept_bits())
                self.last_update = action_update
            else:
                # Undo an action.
                self.context_tree.revert_history(self.environment.action_bits())
                self.last_update = percept_update
            # end if
        # end while

        # Revert agent parameters.
        self.age          = undo_instance.age
        self.total_reward = undo_instance.total_reward
        self.last_update  = undo_instance.last_update
    # end def

    def model_size(self):
        """ Returns the size of the agent's model.

            (Called `modelSize` in the C++ version.)
        """
        return self.context_tree.size()
    # end def

    def model_update_action(self, action):
        """ Update the agent's model of the world with a percept from the
            environment.

            - `observation`: the observation that was received.
            - `reward`: the reward that was received.
        """

        # The last update must have been a percept, else this action update is invalid.
        assert self.environment.is_valid_action(action), "Invalid action given."
        assert self.last_update == percept_update, "Can only perform an action update after a percept update."

        # Update the agent's internal model of the world after performing an action.

        # Get the symbols that represent this action.
        action_symbols = self.encode_action(action)

        # Update the context tree.
        self.context_tree.update_history(action_symbols);

        # Update other properties.
        self.age += 1;
        self.last_update = action_update
    # end def

    def model_update_percept(self, observation, reward):
        """ Update the agent's model of the world with a percept from the
            environment.

            - `observation`: the observation that was received.
            - `reward`: the reward that was received.
        """

        # The last update must have been an action, else this percept update is invalid.
        assert self.last_update == action_update, "Can only perform a percept update after an action update."

        # Update the internal model after performing a percept.

        # Get the symbols that represent this percept from the given observation and reward.
        percept_symbols = self.encode_percept(observation, reward)

        # Are we still meant to be learning?
        if ((self.learning_period > 0) and (self.age > self.learning_period)):
            # No. Update, but don't learn.
            self.context_tree.update_history(percept_symbols)
        else:
            # Yes. Update and learn.
            self.context_tree.update(percept_symbols)
        # end if

        # Update other properties.
        self.total_reward += reward
        self.last_update = percept_update
    # end def

    def percept_probability(self, observation, reward):
        """ Returns the probability of receiving a particular percept
            (the given observation and reward) according to the agent's environment model.

            - `observation`: the observation part of the percept we wish to find the likelihood of.

            - `reward`: the reward part of the percept we wish to find the likelihood of.

            (Called `perceptProbability` in the C++ version.)
        """

        assert self.last_update == action_update, \
               "Can only get perception probabilities after an action update."

        # Encode the given percept.
        percept_symbols = self.encode_percept(observation, reward)

        # Get the probability of seeing this percept using the context tree 
        # to make a prediction.
        return self.context_tree.predict(percept_symbols)
    # end def

    def playout(self, horizon):
        """ Simulate agent/enviroment interaction for a specified amount of steps
            (the given horizon value) where the agent actions are chosen uniformly
            at random and percepts are generated.

            Returns the total reward from the simulation.

            - `horizon`: the number of complete action/percept steps
                         (the search horizon) to simulate.
        """

        total_reward = 0.0

        # Perform `horizon` number of randomly chosen actions.
        for i in xrange(0, horizon):
            # Execute an action chosen uniformly at random.
            action = self.generate_random_action()
            self.model_update_action(action)

            # Sample a percept.
            observation, reward = self.generate_percept_and_update()

            # Increase the total reward by the reward discovered.
            total_reward += reward
        # end while

        # Return the total reward discovered.
        return total_reward
    # end def

    def reset(self):
        """ Resets the agent and clears the context tree.
        """

        # Reset the context tree.
        self.context_tree.clear()

        # Reset the basic agent details.
        agent.Agent.reset(self)
    # end def

    def search(self):
        """ Returns the best action for this agent as determined using the Monte-Carlo Tree Search
            (predictive UCT).
        """

        # Use rhoUCT to search for the next action.

        # Save the agent's current state.
        undo_instance = MC_AIXI_CTW_Undo(self)

        # Create a new search tree.
        search_tree = monte_carlo_search_tree.MonteCarloSearchNode(decision_node)

        # Sample `self.mc_simulations` number of times from the current agent, reverting after each sample.
        for i in xrange(0, self.mc_simulations):
            # Sample from the clone, up to the current horizon
            search_tree.sample(self, self.horizon)

            # Revert the sampling.
            self.model_revert(undo_instance)
        # end for

        # Determine the best action using the tree constructed during sampling,
        # by choosing the action branch from this tree that provides the best expected reward.
        best_action = self.generate_random_action()
        best_mean = -1

        for action in self.environment.valid_actions:
            # Skip this action if the search tree never encountered it.
            if action not in search_tree.children:
                continue
            # end if

            # Get the mean chance of this action, plus a small fudge factor to
            # encourage occasional exploration of other paths.
            mean = search_tree.children[action].mean + (random.random() * 0.0001)

            # Is the mean of this action better than that we've seen so far?
            if mean > best_mean:
                # Yes. Update the best mean and action.
                best_mean = mean
                best_action = action
            # end def
        # end for

        # Return the best action discovered.
        return best_action
    # end def
# end class


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
A script for running AIXI-based agents in an environment, as configured by given options or the
given configuration file.

Usage: python aixi.py [-a | --agent <agent module name>]
                      [-d | --explore-decay <exploration decay value, between 0 and 1>]
                      [-e | --environment <environment module name>]
                      [-h | --agent-horizon <search horizon>]
                      [-l | --learning-period <cycle count>]
                      [-m | --mc-simulations <number of simulations to run each step>]
                      [-o | --option <extra option name>=<value>]
                      [-p | --profile]
                      [-r | --terminate-age <number of cycles before stopping the run>]
                      [-t | --ct-depth <maximum depth of predicting context tree>]
                      [-x | --exploration <exploration factor, greater than 0>]
                      [-v | --verbose]
                      [<environment configuration file name to load>]
"""

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import sys

try:
   import ConfigParser as configparser
except:
   import configparser
# end try

try:
    import cProfile as profile
except:
    import profile
# end try

try:
    import cStringIO as StringIO
except:
    import io as StringIO
# end try

import datetime
import getopt
import inspect
import logging
import random
import sys
import time

# Insert this module's directory (and the 'agents' and 'environments' subdirectory)
# into the system search path, so that agents and environments can be imported by name.
PROJECT_ROOT = os.path.dirname(__file__)
sys.path.insert(0, os.path.join(PROJECT_ROOT, os.pardir))
sys.path.insert(0, os.path.join(PROJECT_ROOT, "agents"))
sys.path.insert(0, os.path.join(PROJECT_ROOT, "environments"))

import pyaixi
from pyaixi import agent, agents, environment, environments, util
from agent import Agent
from agents import *
from environment import Environment
from environments import *

def interaction_loop(agent = None, environment = None, options = {}):
    """ The main agent/environment interaction loop.

        Each interaction cycle begins with the agent receiving an
        observation and reward from the environment.

        Subsequently, the agent selects an action and informs the environment.

        The interactions that took place are logged to the logger.
        When the cycle equals a power of two, a summary of the interactions is printed to
        the standard output.

        - `agent`: the agent object.
        - `environment`: the environment object.
        - `options`: the configuration options.

        (Called `mainLoop` in the C++ version.)
    """

    # Apply a random seed (Default: None)
    random.seed(options.get("random-seed", None))

    # Verbose output (Default: False)
    verbose = bool(options.get("verbose", False))

    # Determine exploration options. (Default: don't explore, don't decay.)
    explore_rate = float(options.get("exploration", 0.0))
    explore = (explore_rate > 0)
    explore_decay = float(options.get("explore-decay", 1.0))
    assert 0.0 <= explore_rate
    assert 0.0 <= explore_decay and explore_decay <= 1.0

    # Determine termination age. (Default: don't terminate)
    terminate_age = int(options.get("terminate-age", 0))
    terminate_check = (terminate_age > 0)
    assert 0 <= terminate_age

    # Determine the cycle after which the agent stops learning (if ever.)
    learning_period = int(options.get("learning-period", 0))
    assert 0 <= learning_period

    # Agent/environment interaction loop.
    cycle = 1
    while not environment.is_finished:
        # Check for agent termination.
        if terminate_check and agent.age > terminate_age:
            break
        # end if

        # Save the current time to compute how long this cycle took.
        cycle_start = datetime.datetime.now()

        # Get a percept from the environment.
        observation = environment.observation
        reward = environment.reward

        # If we're outside the learning period, stop exploring.
        if learning_period > 0 and cycle > learning_period:
            explore = False
        # end if

        # Update the agent's environment model with the new percept.
        agent.model_update_percept(observation, reward)

        # Determine best exploitive action, or explore.
        explored = False
        if explore and (random.random() < explore_rate):
            # Yes, we're still exploring.
            # Generate a random action to explore.
            explored = True
            action = agent.generate_random_action()
        else:
            # No, we're not still exploring.
            # Exploit our past learning to work out the best action.
            action = agent.search()
        # end def

        # Send the action to the environment.
        environment.perform_action(action)

        # Update the agent's environment model with the chosen action.
        agent.model_update_action(action)

        # Calculate how long this cycle took.
        time_taken = datetime.datetime.now() - cycle_start

        # Log this cycle.
        message = "%d, %s, %s, %s, %s, %f, %d, %f, %s, %d" % \
                  (cycle, str(observation), str(reward),
                   str(action), str(explored), explore_rate,
                   agent.total_reward, agent.average_reward(),
                   str(time_taken), agent.model_size())
        print(message)

        # Print to standard output when cycle == 2^n or on verbose option.
        if verbose or (cycle & (cycle - 1)) == 0:
            message = "cycle: %s" % str(cycle) + os.linesep + \
                      "average reward: %f" % agent.average_reward()
            if explore:
                message += os.linesep + "explore rate: %f" % float(explore_rate) + os.linesep
            # end if

            print(message)
        # end def

        # Print environment state if verbose option is true.
        if verbose:
              print(environment.print())
        # end if

        # Update exploration rate.
        if explore:
            explore_rate *= explore_decay
        # end def

        # Update the cycle count.
        cycle += 1
    # end while

    # Print summary to standard output.
    message = "SUMMARY:" + os.linesep + \
              "agent age: %d" % agent.age + os.linesep + \
              "average reward: %f" % agent.average_reward()

    print(message)
# end def

def main(argv):
    """ Entry point of the program. Sets up logging, default configuration values,
        environment and agent before starting the agent/environment interaction cycle
        by calling `interaction_loop`.

        If invalid arguments or options are given, it prints usage help information
        to the standard output and exits.
    """

    # Define some default configuration values.
    default_options = {}
    default_options["agent"]           = "mc_aixi_ctw"
    default_options["agent-horizon"]   = 5
    default_options["ct-depth"]        = 30
    default_options["environment"]     = "coin_flip"
    default_options["exploration"]     = 0.0    # Do not explore.
    default_options["explore-decay"]   = 1.0    # Exploration rate does not decay.
    default_options["learning-period"] = 0      # Learn forever.
    default_options["mc-simulations"]  = 300
    default_options["profile"]         = False  # Whether to profile code.
    default_options["terminate-age"]   = 0      # Never die.
    default_options["verbose"]         = False

    command_line_options = {}

    # Process the command line options and arguments.
    try:
        opts, args = getopt.gnu_getopt(
                                       argv,
                                       'd:e:h:l:m:o:pr:t:vx:',
                                       ['explore-decay=', 'environment=', 'agent-horizon=',
                                        'learning-period=', 'mc-simulations=', 'option', 'profile',
                                        'terminate-age=', 'ct-depth=', 'verbose', 'exploration=',]
                                      )

        for opt, arg in opts:
            if opt == '--help':
                usage()
            # end if
            if opt in ('-d', '--explore-decay'):
                command_line_options["explore-decay"] = float(arg)
                continue
            # end if
            if opt in ('-e', '--environment'):
                command_line_options["environment"] = str(arg)
                continue
            # end if
            if opt in ('-h', '--agent-horizon'):
                command_line_options["agent-horizon"] = int(arg)
                continue
            # end if
            if opt in ('-l', '--learning-period'):
                command_line_options["learning-period"] = int(arg)
                continue
            # end if
            if opt in ('-m', '--mc-simulations'):
                command_line_options["mc-simulations"] = int(arg)
                continue
            # end if
            if opt in ('-o', '--option'):
                # Split the associated argument into a key and value pair, splitting on the '=' symbol.
                parts = arg.split("=")

                # Do we have enough parts to make a key=value pair?
                if len(parts) > 1:
                    key = parts[0].strip()
                    value = '='.join(parts[1:])
                    command_line_options[key] = value
                else:
                    # No. Show the usage, after printing an explantory message.
                    print("Extra option '-o %s' is invalid. " % str(arg) + \
                          "This needs to be in '-o key=value' format." % str(arg))
                    usage()
                # end if
                continue
            # end if
            if opt in ('-p', '--profile'):
                command_line_options["profile"] = True
                continue
            # end if
            if opt in ('-r', '--terminate-age'):
                command_line_options["terminate-age"] = int(arg)
                continue
            # end if
            if opt in ('-t', '--ct-depth'):
                command_line_options["ct-depth"] = int(arg)
                continue
            # end if
            if opt in ('-v', '--verbose'):
                command_line_options["verbose"] = True
                continue
            # end if
            if opt in ('-x', '--exploration'):
                command_line_options["exploration"] = float(arg)
                continue
            # end if
        # end for
    except getopt.GetoptError as e:
        # We got an incorrect option. Show the usage and exit.
        usage()
    # end try

    # Do we have any arguments left over?
    if len(args) > 0:
        # Yes. The first should be the name of a configuration file.
        filename = args[0]

        # Is this a valid filename?
        if not os.path.exists(filename):
            print("Expected argument '%s' to be a configuration filename." % str(filename))
            usage()
        # end if

        # If we're here, we've got a valid filename.
        # Try reading it in as configuration file.
        config_contents = open(filename, 'r').read()

        # Does the configuration contents contain an 'environment' section?
        if config_contents.find("[environment]") == -1:
            # No. Add one to the beginning.
            config_contents = "[environment]" + os.linesep + config_contents
        # end if

        # Convert the contents into an in-memory file-like object, for parsing.
        config_stringio = StringIO.StringIO(config_contents)

        # Parse the given options, giving the default options as defaults to the parser.
        config = configparser.RawConfigParser(default_options)
        config.readfp(config_stringio)

        # Get the configuration options read in as a dictionary.
        # (This should exist in a section called 'environment'.)
        options = dict(config.items('environment'))
    else:
        # No. So set the options to be the default options.
        options = default_options
    # end if

    # Let the command line options override the options read from the configuration file--or
    # the default values--whichever way we got to this point.
    options.update(command_line_options)

    # Print the options we've received, if we've been requested to be verbose.
    verbose = bool(options.get("verbose", False))
    if verbose:
        for option_name, option_value in options.items():
            print("OPTION: '%s' = '%s'" % (str(option_name), str(option_value)))
        # end for
    # end if

    # Print an initial message header.
    message = "cycle, observation, reward, action, explored, " + \
              "explore_rate, total reward, average reward, time, model size"
    print(message)

    # Try to import an agent module with the given name.
    # Check if we've been given a valid agent value.
    agent_name = options["agent"]
    try:
        agent_module = __import__(agent_name, globals(), locals(), [], -1)
    except Exception as e:
        # Exit with an error.
        sys.stderr.write("ERROR: loading agent module '%s' caused error '%s'. Exiting." % \
                         (str(agent_name), str(e)) + os.linesep)
        sys.exit(1)
    # end try

    # Find a subclass of the Agent class in the given module.
    agent_class = None
    agent_classname = ""
    for name in dir(agent_module):
        obj = getattr(agent_module, name)
        if inspect.isclass(obj) and 'Agent' in [cls.__name__ for cls in obj.__bases__]:
            agent_class = obj
            agent_classname = name
            break
        # end if
    # end for

    # Did we find a subclass of Agent?
    if agent_class is None:
        # No. Exit with an error.
        sys.stderr.write("ERROR: agent module '%s' does not contain " % str(agent_name) + \
                         "a valid AIXI agent subclass. (Got '%s' instead.) Exiting." % \
                         str(agent_classname) + os.linesep)
        sys.exit(1)
    # end if

    # Try to import an environment module with the given name.
    # Check if we've been given a valid environment value.
    environment_name = options["environment"]
    environment_classname = ""
    try:
        environment_module = __import__(environment_name, globals(), locals(), [], -1)
    except Exception as e:
        # Exit with an error.
        sys.stderr.write("ERROR: loading environment module '%s' caused error '%s'. Exiting." % \
                         (str(environment_name), str(e)) + os.linesep)
        sys.exit(1)
    # end try

    # Find a subclass of the Environment class in the given module.
    environment_class = None
    for name, obj in inspect.getmembers(environment_module):
        if hasattr(obj, "__bases__") and 'Environment' in [cls.__name__ for cls in obj.__bases__]:
            environment_class = obj
            environment_classname = name
            break
        # end if
    # end for

    # Did we find a subclass of Environment?
    if environment_class is None:
        # No. Exit with an error.
        sys.stderr.write("ERROR: environment module '%s' does not contain " % str(environment_name) + \
                         "a valid AIXI environment subclass. (Got '%s' instead.) Exiting." % \
                          str(environment_classname) + os.linesep)
        sys.exit(1)
    # end if

    # Create an instance of the environment, using the discovered options.
    environment = environment_class(options = options)

    # Copy environment-dependent configuration options to the options.
    options["action-bits"] = environment.action_bits()
    options["observation-bits"] = environment.observation_bits()
    options["percept-bits"] = environment.percept_bits()
    options["reward-bits"] = environment.reward_bits()
    options["max-action"] = environment.maximum_action()
    options["max-observation"] = environment.maximum_observation()
    options["max-reward"] = environment.maximum_reward()

    # Set up the agent, using the created environment, and the updated options.
    agent = agent_class(environment = environment, options = options)

    # Run the main agent/environment interaction loop, profiling if requested to do so.
    if bool(options.get("profile", False)):
        profile.runctx('interaction_loop(agent = agent, environment = environment, options = options)',
                       globals(), locals())
    else:
        interaction_loop(agent = agent, environment = environment, options = options)
    # end def
# end def

def usage():
    """ Prints usage information.
    """

    message = "Usage: python aixi.py [-a | --agent <agent module name>" + os.linesep + \
              "                      [-d | --explore-decay <exploration decay value, between 0 and 1>]" + os.linesep + \
              "                      [-e | --environment <environment module name>]" + os.linesep + \
              "                      [-h | --agent-horizon <search horizon>]" + os.linesep + \
              "                      [-l | --learning-period <cycle count>]" + os.linesep + \
              "                      [-m | --mc-simulations <number of simulations to run each step>]" + os.linesep + \
              "                      [-o | --option <extra option name>=<value>]" + os.linesep + \
              "                      [-p | --profile]" + os.linesep + \
              "                      [-r | --terminate-age <number of cycles before stopping the run>]" + os.linesep + \
              "                      [-t | --ct-depth <maximum depth of predicting context tree>]" + os.linesep + \
              "                      [-x | --exploration <exploration factor, greater than 0>]" + os.linesep + \
              "                      [-v | --verbose]" + os.linesep + \
              "                      [<configuration file name to load>]" + os.linesep + os.linesep

    sys.stderr.write(message)
    sys.exit(2)
# end def


# Start the main function if this file has been executed, and not just imported.
if __name__ == "__main__":
    main(sys.argv[1:])
# end def


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Defines an environment for Kuhn Poker: a simplified, zero-sum version of poker.
"""

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import random

from pyaixi import environment, util

# Define a enumeration to represent agent interactions with the environment,
# such as betting or passing.
kuhn_poker_action_enum = util.enum('aBet', 'aPass')

# Define a enumeration to represent environment observations, such as card values,
# and opponent bet status
# The final observation is of the form `agent-card + opponent-bet-status`.
kuhn_poker_observation_enum = util.enum(oJack = 0, oQueen = 1, oKing = 2, oBet = 0, oPass = 4)

# Define a enumeration to represent rewards as a result of actions: betting and losing,
# betting and winning, passing and losing, passing and winning.
kuhn_poker_reward_enum = util.enum(rBetLoss = 0, rPassLoss = 1, rPassWin = 3, rBetWin = 4)

# Define some shorthand notation for ease of reference.
aBet   = kuhn_poker_action_enum.aBet
aPass  = kuhn_poker_action_enum.aPass

oJack  = kuhn_poker_observation_enum.oJack
oQueen = kuhn_poker_observation_enum.oQueen
oKing  = kuhn_poker_observation_enum.oKing
oBet   = kuhn_poker_observation_enum.oBet
oPass  = kuhn_poker_observation_enum.oPass

rBetLoss  = kuhn_poker_reward_enum.rBetLoss
rPassLoss = kuhn_poker_reward_enum.rPassLoss
rPassWin  = kuhn_poker_reward_enum.rPassWin
rBetWin   = kuhn_poker_reward_enum.rBetWin

class KuhnPoker(environment.Environment):
    """ Kuhn Poker is a simplified, zero-sum, two player poker variant that uses a
        deck of three cards: a King, Queen and Jack.

        While considerably less sophisticated than popular poker variants such as
        Texas Hold'em, well-known strategic concepts such as bluffing and slow-playing
        remain characteristic of strong play.

        In this setup, the agent acts second in a series of rounds. Two actions, pass
        (`aPass`) or bet (`aBet`), are available to each player.

        A bet action requires the player to put an extra chip into play. At the
        beginning of each round, each player puts a chip into play. The environment
        (opponent) then decides whether to pass or bet; betting will win the round
        if the agent subsequently passes, otherwise a showdown will occur.

        In a showdown, the player with the highest card wins the round (i.e. King beats
        Queen, Queen beats Jack). If the environment (opponent) passes, the agent
        can either bet or pass; passing leads immediately to a showdown, while
        betting requires the environment (opponent) to either bet to force a
        showdown, or to pass and let the agent win the round uncontested.

        The winner of the round gains a reward equal to the total chips in play
        (`rPassWin` or `rBetWin`), the loser receives a penalty equal to the number
        of chips they put into play this round (`rPassLoss` or `rBetLoss`).

        At the end of the round, all chips are removed from play and another round begins.

        Domain characteristics:

        - environment: "kuhn_poker"
        - maximum action: 1 (1 bit)
        - maximum observation: 6 (3 bits)
        - maximum reward: 4 (3 bits)
    """

    # Instance attributes.

    # Betting constants.
    bet_probability_king  = 0.7
    bet_probability_queen = (1.0 + bet_probability_king) / 3.0
    bet_probability_jack  = bet_probability_king / 3.0

    # Instance methods.

    def __init__(self, options = {}):
        """ Construct the Kuhn Poker environment from the given options.

             - `options` is a dictionary of named options and their values.
        """

        # Set up the base environment.
        environment.Environment.__init__(self, options = options)

        # Define the acceptable action values.
        self.valid_actions = kuhn_poker_action_enum.keys()

        # Define the acceptable observation values.
        self.valid_observations = kuhn_poker_observation_enum.keys()

        # Define the acceptable reward values.
        self.valid_rewards = kuhn_poker_reward_enum.keys()

        # Set the initial reward.
        self.reward = 0

        # Set up the game.
        self.reset()
    # end def

    def card_to_string(self, card):
        """ Returns a human-readable string version of the given card observation.

            (Called `cardToString` in the C++ version.)
        """

        if card == oJack:
            return "jack"
        elif card == oQueen:
            return "queen"
        elif card == oKing:
            return "king"
        else:
            return ""
    # end def

    def perform_action(self, action):
        """ Receives the agent's action and calculates the new environment percept.
            (Called `performAction` in the C++ version.)
        """

        assert self.is_valid_action(action)

        # Save the action.
        self.action = action

        # If the agent did not call the environments bet then the agent loses
        if self.action == aPass and self.env_action == aBet:
            self.reward = rPassLoss
            self.reset()
            return (self.observation, self.reward)
        # end if

        # If the environment passed and the agent bet, then the environment has
        # a chance to change its mind.
        if self.action == aBet and self.env_action == aPass:
            if self.env_card == oQueen and random.random() < self.bet_probability_queen:
                # Bet with the internal-default probability on seeing a king while having a queen.
                self.env_action = aBet
            elif self.env_card == oKing:
                # The environment always bets if it has a king.
                self.env_action = aBet
            else:
                # Environment continues to pass, so agent wins.
                self.reward = rPassWin
                self.reset()
                return (self.observation, self.reward)
            # end if
        # end if

        # Players have bet the same amount, winner has highest card.
        agent_wins = (self.env_card == oJack) or \
                     (self.env_card == oQueen and self.agent_card == oKing)

        if agent_wins:
            self.reward = rBetWin if self.env_action == aBet else rPassWin
        else:
            self.reward = rBetLoss if self.action == aBet else rPassLoss
        # end if
        self.reset()

        return (self.observation, self.reward)
    # end def

    def print(self):
        """ Returns a string indicating the status of the environment.
        """

        # Add cards and bets to the output.
        message = "agent card = %s" % self.card_to_string(self.agent_previous_card) + \
                  ", environment card = %s" % self.card_to_string(self.env_previous_card) + \
                  ", agent %s" % ("passes" if self.action == aPass else "bets") + \
                  ", environment %s" % ("passes" if self.env_previous_action == aPass else "bets") + \
                  os.linesep

        # Add the winner and the reward to the output.
        agent_wins = (self.reward == rPassWin or self.reward == rBetWin)
        message += "agent %s" % ("wins" if agent_wins else "loses") + \
                   ", reward = %d (%d)" % (self.reward, self.reward - 2) + os.linesep

        return message
    # end def

    def random_card(self):
        """ Returns a card uniformly at random.

            (Called `randomCard` in the C++ version.)
        """

        return random.choice([oJack, oQueen, oKing])
    # end def

    def reset(self):
        """ Resets the environment by randomly placing the tiger behind one door,
            the gold behind the other, and reseating the agent.
        """

        # Save the previous actions/cards for use by print().
        self.env_previous_action = getattr(self, 'env_action', None)
        self.agent_previous_card = getattr(self, 'agent_card', None)
        self.env_previous_card   = getattr(self, 'env_card', None)

        # Deal cards.
        self.agent_card = self.random_card()
        self.env_card = self.agent_card
        while (self.env_card == self.agent_card):
            self.env_card = self.random_card()
        # end def

        # Choose the environment's first action. Bet with a certain probability
        # on jack and king, pass on queen.
        if (self.env_card == oJack):
            self.env_action = aBet if (random.random() < self.bet_probability_jack) else aPass
        elif(self.env_card == oQueen):
            # Always pass on a Queen.
            self.env_action = aPass
        elif(self.env_card == oKing):
            self.env_action = aBet if (random.random() < self.bet_probability_king) else aPass
        # end if

        # Compute an observation: agent-card + environment-bet-status
        self.observation = self.agent_card + (oPass if (self.env_action == aPass) else oBet)
    # end def
# end class


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Define classes to implement context trees according to the Context Tree Weighting algorithm.
"""

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import math
import random

# The value ln(0.5).
# This value is used often in computations and so is made a constant for efficiency reasons.
log_half = math.log(0.5)

class CTWContextTreeNode:
    """ The CTWContextTreeNode class represents a node in an action-conditional context tree.


        The purpose of each node is to calculate the weighted probability of observing
        a particular bit sequence.

        In particular, denote by `n` the current node, by `n0` and `n1`  the child nodes,
        by `h_n` the subsequence of the history relevant to node `n`, and by `a`
        and `b` the number of zeros and ones in `h_n`.

        Then the weighted block probability of observing `h_n` at node `n` is given by

          P_w^n(h_n) :=
  
            Pr_kt(h_n)                        (if n is a leaf node)
            1/2 Pr_kt(h_n) +
            1/2 P_w^n0(h_n0) P_w^n1(h_n1)     (otherwise)

        where `Pr_kt(h_n) = Pr_kt(a, b)` is the Krichevsky-Trofimov (KT) estimator defined by the relations

          Pr_kt(a + 1, b) = (a + 1/2)/(a + b + 1) Pr_kt(a, b)

          Pr_kt(a, b + 1) = (b + 1/2)/(a + b + 1) Pr_kt(a, b)

        and the base case

          Pr_kt(0, 0) := 1


        In both relations, the fraction is referred to as the update multiplier and corresponds to the
        probability of observing a zero (first relation) or a one (second relation) given we have seen
        `a` zeros and `b` ones.

        Due to numerical issues, the implementation uses logarithmic probabilities
        `log(P_w^n(h_n))` and `log(Pr_kt(h_n)` rather than normal probabilities.

        These probabilities are recalculated during updates (`update()`)
        and reversions (`revert()`) to the context tree that involves the node.

        - The KT estimate is accessed and stored using `log_kt`.
          It is updated from the previous estimate by multiplying with the update multiplier as
          calculated by `log_kt_multiplier()`.

        - The weighted probability is access and stored using `log_probability`.
          It is recalculated by `update_log_probability()`.

        In order to calculate these probabilities, `CTWContextTreeNode` also stores:

        - Links to child nodes: `children`

        - The number of symbols (zeros and ones) in the history subsequence relevant to the
          node: `symbol_count`.


        The `CTWContextTreeNode` class is tightly coupled with the `ContextTree` class.

        Briefly, the `ContextTree` class:

        - Creates and deletes nodes.

        - Tells the appropriate nodes to update/revert their probability estimates.

        - Samples actions and percepts from the probability distribution specified
          by the nodes.
    """

    # Instance methods.

    def __init__(self, tree = None):
        """ Construct a node of the context tree.
        """

        # The children of this node.
        self.children = {}

        # The tree object associated with this node.
        self.tree = tree

        # The cached KT estimate of the block log probability for this node.
        # This value is computed only when the node is changed by the update or revert methods.
        # (Called `m_log_kt` in the C++ version.)
        self.log_kt = 0.0

        # The cached weighted log probability for this node.
        # This value is computed only when the node is changed by the update or revert methods.
        # (Called `m_log_probability` and accessed via `logBlockProbability` in the C++ version.)
        self.log_probability = 0.0

        # The count of the symbols in the history subsequence relevant to this node.
        # (Called `m_count` in the C++ version.)
        self.symbol_count = {0: 0, 1: 0}
    # end def

    def is_leaf_node(self):
        """ Return True if the node is a leaf node, False otherwise.
        """

        # If this node has no children, it's a leaf node.
        return self.children == {}
    # end def

    def log_kt_multiplier(self, symbol):
        """ Returns the logarithm of the KT-estimator update multiplier.

           The log KT estimate of the conditional probability of observing a zero given
           we have observed `a` zeros and `b` ones at the current node is

             log(Pr_kt(0 | 0^a 1^b)) = log((a + 1/2)/({a + b + 1))

           Similarly, the estimate of the conditional probability of observing a one is

             log(\Pr_kt(1 |0^a 1^b)) = log((b + 1/2)/(a + b + 1))

           - `symbol`: the symbol for which to calculate the log KT estimate of
             conditional probability.

             0 corresponds to calculating `log(Pr_kt(0 | 0^a 1^b)` and
             1 corresponds to calculating `log(Pr_kt(1 | 0^a 1^b)`.
        """

        numerator = self.symbol_count[symbol] + 0.5
        denominator = self.visits() + 1

        return math.log(numerator / denominator)
    # end def

    def revert(self, symbol):
        """ Reverts the node to its state immediately prior to the last update.
            This involves updating the symbol counts, recalculating the cached
            probabilities, and deleting unnecessary child nodes.

            - `symbol`: the symbol used in the previous update.
        """

        # Decrease the count for this symbol.
        symbol = int(symbol)
        this_symbol_count = self.symbol_count[symbol]
        if this_symbol_count > 1:
            self.symbol_count[symbol] = this_symbol_count - 1
        else:
            self.symbol_count[symbol] = 0
        # end if

        # If the number of visits to the child associated with this symbol is now zero,
        # this child is now redundant, and should be removed.
        redundant_child = self.children.get(symbol, None)
        if redundant_child is not None and (redundant_child.visits() == 0):
            # Decrease the tree size by the size of the children of the redundant child.
            self.tree.tree_size -= redundant_child.size()

            del self.children[symbol]
            del redundant_child
        # end if

        # Revert the KT estimate.
        self.log_kt -= self.log_kt_multiplier(symbol)

        # Update the weighted probability.
        self.update_log_probability()
    # end def

    def size(self):
        """ The number of descendants of this node.
        """

        # Iterate over the direct children of this node, collecting the size of each sub-tree.
        return 1 + sum([child.size() for child in self.children.values()])
    # end def

    def update(self, symbol):
        """ Updates the node after having observed a new symbol.
            This involves updating the symbol counts and recalculating the cached probabilities.

            - `symbol`: the symbol that was observed.
        """

        # Update the KT estimate.
        self.log_kt += self.log_kt_multiplier(symbol)

        # Update the weighted probability.
        self.update_log_probability()

        # Update the count for this symbol.
        self.symbol_count[symbol] += 1
    # end def

    def update_log_probability(self):
        """ This method calculates the logarithm of the weighted probability for this node.

            Assumes that `log_kt` and `log_probability` is correct for each child node.

              log(P^n_w) :=
                  log(Pr_kt(h_n)            (if n is a leaf node)
                  log(1/2 Pr_kt(h_n)) + 1/2 P^n0_w x P^n1_w)
                                            (otherwise)
            and stores the value in log_probability.
     
            Because of numerical issues, the implementation works directly with the
            log probabilities `log(Pr_kt(h_n)`, `log(P^n0_w)`,
            and `log(P^n1_w)` rather than the normal probabilities.

            To compute the second case of the weighted probability, we use the identity

                log(a + b) = log(a) + log(1 + exp(log(b) - log(a)))       a,b > 0

            to rearrange so that logarithms act directly on the probabilities:

                log(1/2 Pr_kt(h_n) + 1/2 P^n0_w P^n1_w) =

                    log(1/2) + log(Pr_kt(h_n))
                      + log(1 + exp(log(P^n0_w) + log(P^n1_w)
                                    - log(Pr_kt(h_n)))

                    log(1/2) + log(P^n0_w) + log(P^n1_w)
                      + log(1 + exp(log(Pr_kt(h_n)
                                           - log(P^n0_w) + log(P^n1_w)))

            In order to avoid overflow problems, we choose the formulation for which
            the argument of the exponent `exp(log(b) - log(a))` is as small as possible.
        """

        # Calculate the log weighted probability.
        # If the current node is a leaf node (i.e. it has no children), this is just the KT estimate.
        # Otherwise, it is an even mixture of the KT estimate, and the product of the
        # weighted probabilities of the children.
        if self.children == {}:
            self.log_probability = self.log_kt
        else:
            # Calculate the sum of the log weighted probabilities of the child nodes.
            log_child_probability = math.fsum([child.log_probability for child in self.children.values()])

            # Calculate the log weighted probability.
            # Use the formulation which has the least chance of overflow.

            # Set 'a' to be the maximum of log_kt and log_child_probability, and 'b' to be the minimum.
            if self.log_kt >= log_child_probability:
                a = self.log_kt
                b = log_child_probability
            else:
                a = log_child_probability
                b = self.log_kt
            # end if

            # Use Python's log1p function to perform `log(1.0 + exp(b - a))`.
            self.log_probability = log_half + a + math.log1p(math.exp(b - a))
        # end if
    # end def

    def visits(self):
        """ Returns the number of times this context has been visited.
            This is the sum of the visits of the (immediate) child nodes.
        """

        return self.symbol_count[0] + self.symbol_count[1]
    # end def
# end class


class CTWContextTree:
    """ The high-level interface to an action-conditional context tree.
        Most of the mathematical details are implemented in the CTWContextTreeNode class, which is used to
        represent the nodes of the tree.
        CTWContextTree stores a reference to the root node of the tree (`root`), the history of
        updates to the tree (`history`), and the maximum depth of the tree (`depth`).

        It is primarily concerned with calling the appropriate functions in the appropriate nodes
        in order to deliver certain functionality:

        - `update(symbol_or_list_of_symbols)` updates the tree and the history
          after the agent has observed new percepts.

        - `update_history(symbol_or_list_of_symbols)` updates just the history
          after the agent has executed an action.

        - `revert()` reverts the last update to the tree.

        - `revert_history()` deletes the recent history.

        - `predict()` predicts the probability of future outcomes.

        - `generate_random_symbols_and_update()` samples a sequence from the
           context tree, updating the tree with each symbol as it is sampled.

        - `generate_random_symbols()` samples a sequence of a specified length,
           updating the tree with each symbol as it is sampled, then reverting all the
           updates so that the tree is in the same state as it was before the
           sampling.
    """

    def __init__(self, depth):
        """ Create a context tree of specified maximum depth.
            Nodes are created as needed.

            - `depth`: the maximum depth of the context tree.
        """

        # An list used to hold the nodes in the context tree that correspond to the current context.
        # It is important to ensure that `update_context()` is called before accessing the contents
        # of this list as they may otherwise be inaccurate.
        self.context = []

        # The maximum depth of the context tree.
        # (Called `m_depth` in the C++ version.)
        assert depth >= 0, "The given tree depth must be greater than zero."
        self.depth = depth

        # The history (a list) of symbols seen by the tree.
        # (Called `m_history` in the C++ version.)
        self.history = []

        # The root node of the context tree.
        # (Called `m_root` in the C++ version.)
        self.root = CTWContextTreeNode(tree = self)

        # The size of this tree.
        self.tree_size = 1
    # end def

    def clear(self):
        """ Clears the entire context tree including all nodes and history.
        """

        # Reset the history.
        self.history = []

        # Set a new root object, and reset the tree size.
        self.root.tree = None
        del self.root
        self.root = CTWContextTreeNode(tree = self)
        self.tree_size = 1

        # Reset the context.
        self.context = []
    # end def

    def generate_random_symbols(self, symbol_count):
        """ Returns a symbol string of a specified length by sampling from the context tree.

            - `symbol_count`: the number of symbols to generate.

            (Called `genRandomSymbols` in the C++ version.)
        """
        symbol_list = self.generate_random_symbols_and_update(symbol_list, symbol_count)
        self.revert(symbol_count)

        return symbol_list
    # end def

    def generate_random_symbols_and_update(self, symbol_count):
        """ Returns a specified number of random symbols distributed according to
            the context tree statistics and update the context tree with the newly
            generated symbols.

            - `symbol_count`: the number of symbols to generate.

            (Called `genRandomSymbolsAndUpdate` in the C++ version.)
        """

        symbol_list = []
        for i in xrange(0, symbol_count):
            # Pick either 0 or 1 based on the probability of the symbol 1 occuring in the context tree.
            symbol = 1 if (random.random() < self.predict(1)) else 0
            symbol_list += [symbol]
            self.update(symbol)
        # end for

        return symbol_list
    # end def

    def predict(self, symbol_list):
        """ Returns the conditional probability of a symbol (or a list of symbols), considering the history.

            Given a history sequence `h` and a symbol `y`, the estimated probability is given by

              rho(y | h) = rho(hy)/rho(h)

            where `rho(h) = P_w^epsilon(h)` is the weighted probability estimate of observing `h`
            evaluated at the root node `epsilon` of the context tree.

            - `symbol_list` The symbol (or list of symbols) to estimate the conditional probability of.
                            0 corresponds to `rho(0 | h)` and 1 to `rho(1 | h)`.
        """

        # Ensure that we have a list, by making this a list if it's a single symbol.
        if type(symbol_list) != list:
            symbol_list = [symbol_list]
        # end if


        # If there is insufficient context for a prediction, return the uniform
        # prediction 0.5 ^ length.
        symbol_list_length = len(symbol_list)
        if ((len(self.history) + symbol_list_length) <= self.depth):
            return 0.5 if symbol_list_length == 1 else math.pow(0.5, symbol_list_length)
        # end if

        # Calculate the probability of the symbol s given the history h using
        # p(s | h) = p(hs) / p(h) = exp(ln p(hs) - ln p(h)).
        prob_history = self.root.log_probability
        self.update(symbol_list)
        prob_sequence = self.root.log_probability
        self.revert(symbol_list_length)

        return math.exp(prob_sequence - prob_history)
    # end def

    def revert(self, symbol_count = 1):
        """ Restores the context tree to its state prior to a specified number of updates.
     
            - `num_symbols`: the number of updates (symbols) to revert. (Default of 1.)
        """

        # Traverse the tree from leaf to root according to the context.
        for i in xrange(0, symbol_count):
            # Check if we have updates to revert.
            if len(self.history) == 0:
                return
            # end if

            # Get the most recent symbol and delete from the history.
            symbol = self.history.pop()

            # Traverse the tree from leaf to root according to the context. Update the
            # probabilities and symbol counts for each node. Delete unnecessary nodes.
            if len(self.history) >= self.depth:
                self.update_context()

                # Step backwards through the nodes in the context in reverse context order.
                # (Only go as deep as the current tree depth, though.)
                for context_node in reversed(self.context[:self.depth]):
                    context_node.revert(symbol)
                # end for
            # end if
        # end for
    # end def

    def revert_history(self, symbol_count = 1):
        """ Shrinks the history without affecting the context tree.

            (Called `revertHistory` in the C++ version.)
        """

        assert symbol_count > 0, "The given symbol count should be greater than 0."
        history_length = len(self.history)
        assert history_length >= symbol_count, "The given symbol count must be greater than the history length."

        new_size = history_length - symbol_count
        self.history = self.history[:new_size]
    # end def

    def size(self):
        """ Returns the number of nodes in the context tree.
        """

        # Return the value stored and updated by the children nodes.
        return self.tree_size
    # end def

    def update(self, symbol_list):
        """ Updates the context tree with a new (binary) symbol, or a list of symbols.
            Recalculates the log weighted probabilities and log KT estimates for each affected node.

            - `symbol_list`: the symbol (or list of symbols) with which to update the tree.
                              (The context tree is updated with symbols in the order they appear in the list.)
        """

        # Ensure that we have a list, by making this a list if it's a single symbol.
        if type(symbol_list) != list:
            symbol_list = [symbol_list]
        # end if

        # Traverse the tree from leaf to root according to the context.
        for symbol in symbol_list:
            # Update the probabilities and symbol counts for each node.
            symbol = int(symbol)
            if len(self.history) >= self.depth:
                self.update_context()

                # Step backwards through the nodes in the context in reverse context order.
                # (Only go as deep as the current tree depth, though.)
                for context_node in reversed(self.context[:self.depth]):
                    context_node.update(symbol)
                # end for
            # end if

            # Add this symbol to the history.
            self.update_history(symbol)
        # end for
    # end def

    def update_context(self):
        """ Calculates which nodes in the context tree correspond to the current
            context, and adds them to `context` in order from root to leaf.

            In particular, `context[0]` will always correspond to the root node
            and `context[self.depth]` corresponds to the relevant leaf node.

            Creates the nodes if they do not exist.

            (Called `updateContext` in the C++ version.)
        """

        # Ensure that the length of the history is greater than or equal to the tree depth for safety.
        assert len(self.history) >= self.depth, "The history length must be greater than the tree depth."

        # Traverse the tree from root to leaf according to the context.
        # Save the path taken and create new nodes as necessary.
        self.context = [self.root]
        node = self.root
        update_depth = 1
        for symbol in reversed(self.history):
            # Find the relevant child node of the current node for the current symbol, if it exists.
            symbol = int(symbol)
            if symbol in node.children:
                node = node.children[symbol]
            else:
                # No child exists for this symbol.

                # Create a new node for the context, and add it into the tree under the current symbol.
                new_node = CTWContextTreeNode(tree = self)
                node.children[symbol] = new_node

                # Increase the size of the tree by 1, for the new node.
                self.tree_size += 1

                # Move onto this new node.
                node = new_node
            # end if

            # Add the node to the context path.
            self.context += [node]

            # Have we hit the end of the update depth yet?
            update_depth += 1
            if update_depth > self.depth:
                # Yes. Stop updating the context.
                break
            # end if
        # end for
    # end def

    def update_history(self, symbol_list):
        """ Appends a symbol (or a list of symbols) to the tree's history without updating the tree.

            - `symbol_list`: the symbol (or list of symbols) to add to the history.

            (Called `updateHistory` in the C++ version.)
        """

        # Ensure that we have a list, by making this a list if it's a single symbol.
        if type(symbol_list) != list:
            symbol_list = [symbol_list]
        # end if

        self.history += symbol_list
    # end def
# end class


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Defines an environment for Tic Tac Toe.
"""

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import random

from pyaixi import environment, util

# Define a enumeration to represent environment observations: either a square
# is empty, filled with the agent's piece, or the environment's piece.
tictactoe_observation_enum = util.enum('oEmpty', 'oAgent', 'oEnv')

# Define a enumeration to represent rewards as a result of actions: invalid actions,
# losses, null, draws, and wins.
# NOTE: since the enumeration values need to be positive, these values are defined relative to 3.
#       So -3 points is 0, -2 points is 1, and 2 points is 5.
tictactoe_reward_enum = util.enum(rInvalid = 0, rLoss = 1, rNull = 3, rDraw = 4, rWin = 5)

# Define some shorthand notation for ease of reference.
oEmpty  = tictactoe_observation_enum.oEmpty
oAgent  = tictactoe_observation_enum.oAgent
oEnv    = tictactoe_observation_enum.oEnv

rInvalid = tictactoe_reward_enum.rInvalid
rLoss    = tictactoe_reward_enum.rLoss
rNull    = tictactoe_reward_enum.rNull
rDraw    = tictactoe_reward_enum.rDraw
rWin     = tictactoe_reward_enum.rWin

class Tic_Tac_Toe(environment.Environment):
    """ In this domain, the agent plays repeated games of TicTacToe against an
        opponent who moves randomly. If the agent wins the game, it receives a
        reward of 2. If there is a draw, the agent receives a reward of 1. A loss
        penalizes the agent by -2. If the agent makes an illegal move, by moving on
        top of an already filled square, then it receives a reward of -3. A legal
        move that does not end the game earns no reward. An illegal reward causes
        the game to restart.

        Domain characteristics:
        - environment: "tictactoe"
        - maximum action: 8 (4 bits)
        - maximum observation: 174672 (18 bits)
          - 174672 (decimal) = 101010101010101010 (binary)
        - maximum reward: 5 (3 bits)
    """

    # Instance methods.

    def __init__(self, options = {}):
        """ Construct the Tic Tac Toe environment from the given options.

             - `options` is a dictionary of named options and their values.
        """

        # Set up the base environment.
        environment.Environment.__init__(self, options = options)

        # Define the acceptable action values.
        self.valid_actions = xrange(0, 9)

        # Define the acceptable observation values.
        self.valid_observations = xrange(0, 174672 + 1)

        # Define the acceptable reward values.
        self.valid_rewards = tictactoe_reward_enum.keys()

        # Set the initial reward.
        self.reward = 0

        # Set up the game.
        self.reset()
    # end def

    def check_win(self):
        """ Check if either player has won the game.
            Returns True if so, False otherwise.

            (Called `checkWin` in the C++ version.)
        """

        # Check if we've got a row of three matching symbols.
        for r in xrange(0, 3):
            # Is this row all matching, non-empty symbols?
            if (self.board[r][0] != oEmpty and \
                self.board[r][0] == self.board[r][1] and \
                self.board[r][1] == self.board[r][2]):
                # Yes. Someone has won.
                return True
            # end if
        # end for

        # Check if we've got any columns of three matching symbols.
        for c in xrange(0, 3):
            # Is this column all matching, non-empty symbols?
            if (self.board[0][c] != oEmpty and \
                self.board[0][c] == self.board[1][c] and \
                self.board[1][c] == self.board[2][c]):
                # Yes. Someone has won.
                return True
            # end if
        # end for

        # Check the diagonals.
        if (self.board[1][1] != oEmpty and \
            self.board[0][0] == self.board[1][1] and \
            self.board[1][1] == self.board[2][2]):
            return True
        # end if

        if (self.board[1][1] != oEmpty and \
            self.board[0][2] == self.board[1][1] and \
            self.board[1][1] == self.board[2][0]):
            return True
        # end if

        # If we're here, there's no winner yet.
        return False
    # end def

    def compute_observation(self):
        """ Encodes the state of each square into an overall observation and saves
            the result in self.observation.
            Each cell corresponds to two bits.

            (Called `computeObservation` in the C++ version.)
        """

        self.observation = 0
        for r in xrange(0, 3):
            for c in xrange(0, 3):
                # Shift the existing observation up by 2 bits, and add the current observation to
                # that.
                self.observation = self.board[r][c] + (4 * self.observation)
            # end for
        # end for
    # end def

    def perform_action(self, action):
        """ Receives the agent's action and calculates the new environment percept.
            (Called `performAction` in the C++ version.)
        """

        assert self.is_valid_action(action)

        # Save the action.
        self.action = action

        # Increment the actions-since-reset counter.
        self.actions_since_reset += 1

        # Decode the action into the desired row and column to update.
        r = int(action / 3)
        c = action % 3

        # If agent makes an invalid move, give the appropriate (lack of) reward and clear the board.
        if (self.board[r][c] != oEmpty):
            self.reward = rInvalid
            self.reset()
            return
        # end def

        # The agent makes their move.
        self.board[r][c] = oAgent

        # If the agent wins or draws, give an appropriate reward and clear the board.
        if (self.check_win()):
            self.reward = rWin
            self.reset()
            return
        elif self.actions_since_reset == 5:
            # If it's been 5 actions since the reset, then this must be a draw, as there's
            # no way to win from this point.
            self.reward = rDraw
            self.reset()
            return
        # end def

        # The environment makes a random play.
        while (self.board[r][c] != oEmpty):
            # Keep picking board positions at random until we find an unoccupied spot.
            r = random.randrange(0, 3)
            c = random.randrange(0, 3)
        # end while

        # If we're here, we've got an unoccupied spot.
        self.board[r][c] = oEnv

        # If the environment has won, give an appropriate reward and clear the board.
        if (self.check_win()):
            self.reward = rLoss
            self.reset()
            return
        # end def

        # The game continues.
        self.reward = rNull
        self.compute_observation()

        return (self.observation, self.reward)
    # end def

    def print(self):
        """ Returns a string indicating the status of the environment.
        """

        # Show what just happened, correcting the reward for being defined relative to 3.
        message = "action = %s, observation = %s, reward = %s (%d), board:" % \
                  (self.action, self.observation, self.reward, (self.reward - 3)) + os.linesep

        # Display the current state of the board.
        for r in xrange(0, 3):
            for c in xrange(0, 3):
                b = self.board[r][c]
                message += "." if b == oEmpty else ("A" if b == oAgent else "O")
            # end for
            message += os.linesep
        # end for
        message += os.linesep

        return message
    # end def

    def reset(self):
        """ Begin a new game.
        """

        # Set up the board.
        self.board = {}

        for r in xrange(0, 3):
            for c in xrange(0, 3):
                # Ensure the row exists.
                if r not in self.board:
                    self.board[r] = {}
                # end if

                # Set this element to be empty.
                self.board[r][c] = oEmpty
            # end for
        # end for

        # Set an initial observation.
        self.compute_observation()

        # Set the actions-since-reset marker.
        self.actions_since_reset = 0
    # end def
# end class


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Defines an environment for an agent interacting with an environment where there's
a tiger and a pot of gold hidden separately, behind two closed doors.
"""

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import random

from pyaixi import environment, util

# Define a enumeration to represent agent interactions with the environment,
# such as listening, or opening doors.
tiger_action_enum = util.enum('aListen', 'aLeft', 'aRight')

# Define a enumeration to represent environment observations: either not
# hearing the tiger, or hearing it from behind either door.
tiger_observation_enum = util.enum('oNull', 'oLeft', 'oRight')

# Define a enumeration to represent rewards as a result of actions: being eaten by the tiger,
# getting information from listening, or finding the gold
# NOTE: since the enumeration values need to be positive, these values are defined relative to
#       100.
#       So -100 points is 0, -1 points is 99, and 10 points is 110.
tiger_reward_enum = util.enum(rEaten = 0, rListen = 99, rGold = 110)

# Define some shorthand notation for ease of reference.
aListen = tiger_action_enum.aListen
aLeft   = tiger_action_enum.aLeft
aRight  = tiger_action_enum.aRight

oNull  = tiger_observation_enum.oNull
oLeft  = tiger_observation_enum.oLeft
oRight = tiger_observation_enum.oRight

rEaten  = tiger_reward_enum.rEaten
rListen = tiger_reward_enum.rListen
rGold   = tiger_reward_enum.rGold

class Tiger(environment.Environment):
    """ The environment dynamics are as follows: a tiger and a pot of gold are
        hidden behind one of two doors.

        Initially the agent starts facing both doors. The agent has a choice of
        one of three actions: listen, open the left door, or open the right door.

        If the agent opens the door hiding the tiger, it suffers a -100 penalty.
        If it opens the door with the pot of gold, it receives a reward of 10.

        If the agent performs the listen action, it receives a penalty of -1 and
        an observation that correctly describes where the tiger is with 0.85 probability.

        Domain characteristics:

        - environment: "tiger"
        - maximum action: 2 (2 bits)
        - maximum observation: 2 (2 bits)
        - maximum reward: 110 (7 bits)

        Configuration options:
        - `tiger-listen-accuracy`: the probability that listening for the
                                   tiger results in the correct observation of the tiger's
                                   whereabouts (`listen_accuracy`).

                                   Must be a floating point number between 0.0
                                   and 1.0 inclusive.

                                   Default value is `default_listen_accuracy`, 85%.
                                   (Optional.)
    """

    # Instance attributes.

    # Set the default probability for the listen accuracy.
    default_listen_accuracy = 0.85

    # Instance methods.

    def __init__(self, options = {}):
        """ Construct the Tiger environment from the given options.

             - `options` is a dictionary of named options and their values.

            The following options in `options` are optional:
             - `tiger-listen-accuracy`: the probability of correctly observing the tiger's location from listening.
                                        Default value: 85%.
        """

        # Set up the base environment.
        environment.Environment.__init__(self, options = options)

        # Define the acceptable action values.
        self.valid_actions = tiger_action_enum.keys()

        # Define the acceptable observation values.
        self.valid_observations = tiger_observation_enum.keys()

        # Define the acceptable reward values.
        self.valid_rewards = tiger_reward_enum.keys()

        # Set the accuracy of the listen action.
        if 'tiger-listen-accuracy' not in options:
            options["tiger-listen-accuracy"] = self.default_listen_accuracy
        # end if
        self.listen_accuracy = float(options["tiger-listen-accuracy"])

        # Make sure the accuracy value is valid.
        assert 0.0 <= self.listen_accuracy and self.listen_accuracy <= 1.0

        # Place the tiger.
        self.place_tiger()

        # Set an initial percept.
        self.observation = oNull
        self.reward = 0
    # end def

    def perform_action(self, action):
        """ Receives the agent's action and calculates the new environment percept.
            (Called `performAction` in the C++ version.)
        """

        assert self.is_valid_action(action)

        # Save the action.
        self.action = action

        if (action == aListen):
            # Listen for the tiger, and return the correct door with probability
            # equal to self.listen_accuracy.
            self.reward = rListen
            self.observation = self.tiger if random.random() < self.listen_accuracy else self.gold
        else:
            # Open a door. Set the reward according to what we find.
            if (action == aLeft):
                self.reward = rEaten if self.tiger == oLeft else rGold
            elif (action == aRight):
                self.reward = rEaten if self.tiger == oRight else rGold
            # end if

            # Set the observation to null, and replace the tiger and the gold.
            self.observation = oNull
            self.place_tiger()
        # end if

        return (self.observation, self.reward)
    # end def

    def place_tiger(self):
        """ Randomly places the tiger behind one door and gold behind the other.

            (Called `placeTiger` in the C++ version.)
        """

        self.tiger = oLeft if random.random() < 0.5 else oRight
        self.gold  = oRight if self.tiger == oLeft else oLeft
    # end def

    def print(self):
        """ Returns a string indicating the status of the environment.
        """

        action_text      = {aListen: "listen",
                            aLeft: "open left door",
                            aRight: "open right door"}
        observation_text = {oNull: "null",
                            oLeft: "hear tiger at left door",
                            oRight: "hear tiger at right door"}
        reward_text      = {rEaten: "eaten",
                            rListen: "listen",
                            rGold: "gold!"}

        # Show what just happened, correcting the reward for being defined relative to 100.
        message = "action = %s, observation = %s, reward = %s (%d)" % \
                  (action_text[self.action],
                   observation_text[self.observation],
                   reward_text[self.reward],
                   (self.reward - 100))

        return message
    # end def
# end class


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Define a class to implement a Monte Carlo search tree.
"""

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import math
import random

from pyaixi import util

# An enumeration type used to specify the type of Monte Carlo search node.
# Chance nodes represent a set of possible observation
# (one child per observation) while decision nodes
# represent sets of possible actions (one child per action).
# Decision and chance nodes alternate.
nodetype_enum = util.enum('chance', 'decision')

# Define some short cuts for ease of reference.
chance_node = nodetype_enum.chance
decision_node = nodetype_enum.decision

class MonteCarloSearchNode:
    """ A class to represent a node in the Monte Carlo search tree.
        The nodes in the search tree represent simulated actions and percepts
        between an agent following an upper confidence bounds (UCB) policy and a generative
        model of the environment represented by a context tree.

        The purpose of the tree is to determine the expected reward of the
        available actions through sampling. Sampling proceeds several time steps
        into the future according to the size of the agent's horizon.
        (`MC_AIXI_CTW_Agent.horizon`)
 
        The nodes are one of two types (`nodetype_enum`), decision nodes are those
        whose children represent actions from the agent and chance nodes are those
        whose children represent percepts from the environment.

        Each MonteCarloSearchNode maintains several bits of information:

          - The current value of the sampled expected reward
            (`MonteCarloSearchNode.mean`, `MonteCarloSearchNode.expectation`).

          - The number of times the node has been visited during the sampling
            (`MonteCarloSearchNode.visits`).

          - The type of the node (MonteCarloSearchNode.type).

          - The children of the node (`MonteCarloSearchNode.children`).
            The children are stored in a dictionary indexed by action (if
            it is a decision node) or percept (if it is a chance node).

        The `MonteCarloSearchNode.sample` method is used to sample from the current node and
        the `MonteCarloSearchNode.selectAction` method is used to select an action according
        to the UCB policy.
    """

    # Class attributes.

    # Exploration constant for the UCB action policy.
    exploration_constant = 2.0

    # Unexplored action bias.
    unexplored_bias = 1000000000.0

    # Instance methods.

    def __init__(self, nodetype):
        """ Create a new search node of the given type.
        """

        # The children of this node.
        # The symbols used as keys at each level may be either action or observation,
        # depending on what type of node this is.
        self.children = {}

        # The sampled expected reward of this node.
        # (Called `m_mean` and `expectation()` in the C++ version.)
        self.mean = 0.0

        # The type of this node indicates whether its children represent actions
        # (decision node) or percepts (chance node).
        # (Called `m_type` in the C++ version.)
        assert nodetype in nodetype_enum, "The given value %s is a not a valid node type." % str(nodetype)
        self.type = nodetype

        # The number of times this node has been visited during sampling.
        # (Called `m_visits` in the C++ version.)
        self.visits = 0
    # end def

    def sample(self, agent, horizon):
        """ Returns the accumulated reward from performing a single sample on this node.

            - `agent`: the agent doing the sampling

            - `horizon`: how many cycles into the future to sample
        """

        # Set an initial reward.
        reward = 0.0

        # Do we need to continue sampling, or use a playout policy?
        # Have we already reached the given horizon or the maximum search depth?
        if (horizon == 0):
            # Yes, we've reached the maximum search depth.
            # Return the initial reward.
            return reward
        elif (self.type == chance_node):
            # We're at a chance node, so we need to continue sampling.

            # Generate a percept at random using the agent's environment model,
            # and continue sampling.
            observation, random_reward = agent.generate_percept_and_update()

            # If this observation is new to this node, add it as a decision observation child node.
            if observation not in self.children:
                self.children[observation] = MonteCarloSearchNode(decision_node)
            # end def
            observation_child = self.children[observation]

            # Get the reward for this observation by continuing the search recursively onto the child.
            reward = random_reward + observation_child.sample(agent, horizon - 1)
        elif (self.visits == 0):
            # We are at an (unvisited) decision node.
            # Either the node is previously unvisited, or we have exceeded the maximum tree depth.
            # Either way, use the playout policy to estimate the future reward.
            reward = agent.playout(horizon)
        else:
            # We are at a previously-visited decision node.
            # Choose an action according to the UCB policy and continue sampling.
            action = self.select_action(agent)
            agent.model_update_action(action)

            # If this action is new to this node, add it as a chance action child node.
            # Do we have a child corresponding to this action?
            if action not in self.children:
                # No. Create one.
                self.children[action] = MonteCarloSearchNode(chance_node)
            # end def
            action_child = self.children[action]

            # Get the reward for this action by continuing the search recursively onto the child.
            # NOTE: should this be horizon - 1, like observation_child_sample above?
            #       (The C++ version is just the horizon as well.)
            reward = action_child.sample(agent, horizon)
        # end def

        # Update the expected reward and number of visits to the current node.
        visits = float(self.visits)
        self.mean = (reward + (visits * self.mean)) / (visits + 1.0)
        self.visits += 1

        # Return the calculated reward.
        return reward
    # end def

    def select_action(self, agent):
        """ Returns an action selected according to UCB policy.

             - `agent`: the agent which is doing the sampling.

            (Called `selectAction` in the C++ version.)
        """

        explore_bias = float(agent.horizon * agent.maximum_reward())
        exploration_numerator = (self.exploration_constant * math.log(self.visits))

        # Compute the best action according to the UCB formula.
        best_action = None
        best_priority = float("-inf")
        for action in agent.environment.valid_actions:
            # Find any children nodes related to this action.
            node = self.children.get(action, None)

            # Use the UCB formula to determine priority of this node.
            priority = 0.0
            if (node is None or node.visits == 0):
                # This is a previously unexplored node.
                # Give it the unexplored bias.
                priority = self.unexplored_bias
            else:
                # This is a previously explored node.
                priority = node.mean + (explore_bias * math.sqrt(exploration_numerator / node.visits))
            # end def

            # Update the best action if necessary, breaking ties randomly.
            if (priority > (best_priority + (random.random() * 0.001))):
                best_action = action
                best_priority = priority
            # end if
        # end for

        return best_action
    # end def
# end class


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Defines an environment for AIXI agents.
"""

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sys
import util

class Environment:
    """ Base class for the various agent environments.

        Each individual environment should inherit from this class and implement the appropriate methods.

        In particular, the constructor should set up the environment as appropriate, including
        setting the initial observation and reward, as well as setting appropriate
        values for the configuration options:
         - `agent-actions`
         - `observation-bits`
         - `reward-bits`

        Following this, the agent and environment interact in a cyclic fashion. The
        agent receives the observation and reward using `Environment.getObservation`
        and `Environment.getReward` before supplying the environment with an action
        via `Environment.performAction`.

        Upon receiving an action, the environment updates the observation and reward.
        At the beginning of each cycle, the value of `Environment::isFinished` is checked.

        If it is true then there is no more interaction between the agent and environment,
        and the program exits. Otherwise the interaction continues indefinitely.
    """

    def __init__(self, options = {}):
        """ Construct an agent environment.
        """

        # Set the current action to null/None.
        # (Called `m_action` and `getAction` in the C++ version.)
        self.action = None

        # Set whether the environment is finished.
        # (Called `isFinished` in the C++ version.)
        self.is_finished = False

        # Set the current observation to null/None.
        # (Called `m_observation` and `getObservation()` in the C++ version.)
        self.observation = None

        # Store the given options.
        self.options = options

        # Set the current reward to null/None.
        # (Called `m_reward` in the C++ version.)
        self.reward = None

        # Defines the acceptable action values.
        self.valid_actions = []

        # Define the acceptable observation values.
        self.valid_observations = []

        # Define the acceptable reward values.
        self.valid_rewards = []
    # end def

    def __unicode__(self):
        """ Returns a string representation of this environment instance.
        """
        return "action = " + str(self.action) + ", observation = " + \
               str(self.observation) + ", reward = " + str(self.reward)
    # end def

    # Make a compatible string function for the current Python version.
    if sys.version_info[0] >= 3:
        # For Python 3.
        def __str__(self):
            return self.__unicode__()
        # end def
    else:
        # For Python 2.
        def __str__(self):
            return self.__unicode__().encode('utf8')
        # end def
    # end def

    def action_bits(self):
        """ Returns the maximum number of bits required to represent an action.
            (Called `actionBits` in the C++ version.)
        """

        # Find the largest sized observation.
        maximum_bits = 0
        for action in self.valid_actions:
            bits_for_this_action = util.bits_required(action)
            if bits_for_this_action > maximum_bits:
                maximum_bits = bits_for_this_action
            # end if
        # end for

        return maximum_bits
    # end def

    def is_valid_action(self, action):
        """ Returns whether the given action is valid.
            (Called `isValidAction` in the C++ version.)
        """
        return action in self.valid_actions
    # end def

    def is_valid_observation(self, observation):
        """ Returns whether the given observation is valid.
            (Called `isValidObservation` in the C++ version.)
        """
        return observation in self.valid_observations
    # end def

    def is_valid_reward(self, reward):
        """ Returns whether the given reward is valid.
            (Called `isValidReward` in the C++ version.)
        """
        return reward in self.valid_rewards
    # end def

    def maximum_action(self):
        """ Returns the maximum possible action.
            (Called `maxAction` in the C++ version.)
        """

        # The largest action is the last in the list of valid actions.
        # Else, it's null/None.
        return self.valid_actions[-1] if len(self.valid_actions) > 0 else None
    # end def

    def maximum_observation(self):
        """ Returns the maximum possible observation.
            (Called `maxObservation` in the C++ version.)
        """

        # The largest observation is the last in the list of valid observations.
        # Else, it's null/None.
        return self.valid_observations[-1] if len(self.valid_observations) > 0 else None
    # end def

    def maximum_reward(self):
        """ Returns the maximum possible reward.
            (Called `maxReward` in the C++ version.)
        """

        # The largest reward is the last in the list of valid rewards.
        # Else, it's null/None.
        return self.valid_rewards[-1] if len(self.valid_rewards) > 0 else None
    # end def

    def minimum_action(self):
        """ Returns the minimum possible action.
            (Called `minAction` in the C++ version.)
        """

        # The smallest action is the first in the list of valid actions.
        # Else, it's null/None.
        return self.valid_actions[1] if len(self.valid_actions) > 0 else None
    # end def

    def minimum_observation(self):
        """ Returns the minimum possible observation.
            (Called `minObservation` in the C++ version.)
        """

        # The smallest observation is the first in the list of valid observations.
        # Else, it's null/None.
        return self.valid_observations[1] if len(self.valid_observations) > 0 else None
    # end def

    def minimum_reward(self):
        """ Returns the minimum possible reward.
            (Called `minReward` in the C++ version.)
        """

        # The smallest reward is the first in the list of valid rewards.
        # Else, it's null/None.
        return self.valid_rewards[1] if len(self.valid_rewards) > 0 else None
    # end def

    def observation_bits(self):
        """ Returns the maximum number of bits required to represent an observation.
            (Called `observationBits` in the C++ version)
        """

        # Find the largest sized observation.
        maximum_bits = 0
        for observation in self.valid_observations:
            bits_for_this_observation = util.bits_required(observation)
            if bits_for_this_observation > maximum_bits:
                maximum_bits = bits_for_this_observation
            # end if
        # end for

        return maximum_bits
    # end def

    def percept_bits(self):
        """ Returns the maximum number of bits required to represent a percept.
            (Called `perceptBits` in the C++ version.)
        """

        return self.observation_bits() + self.reward_bits()
    # end def

    def perform_action(self, action):
        """ Receives the agent's action and calculates the new environment percept.
            (Called `performAction` in the C++ version.)
        """
        # To be overriden by inheriting classes.
        pass
    # end def

    def print(self):
        """ String representation convenience method from the C++ version.
        """
        return str(self)
    # end def

    def reward_bits(self):
        """ Returns the maximum number of bits required to represent a reward.
            (Called `rewardBits` in the C++ version)
        """

        # Find the largest sized reward.
        maximum_bits = 0
        for reward in self.valid_rewards:
            bits_for_this_reward = util.bits_required(reward)
            if bits_for_this_reward > maximum_bits:
                maximum_bits = bits_for_this_reward
            # end if
        # end for

        return maximum_bits
    # end def
# end class



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Defines an environment for a two-dimensional maze.
"""

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import random
import sys

from pyaixi import environment, util

# Define a enumeration to represent agent interactions with the environment,
# such as going left, up, right, or down.
maze_action_enum = util.enum('aLeft', 'aUp', 'aRight', 'aDown')

# Define a enumeration to represent environment observations: either a cell
# is empty, or has a wall in various (bit) positions.
maze_observation_enum = util.enum(oNull = 0, oLeftWall = 1, oUpWall = 2, oRightWall = 4, oDownWall = 8)

# Define an enumber to represent observation encoding constants.
maze_observation_encoding_enum = util.enum('cUninformative', 'cWalls', 'cCoordinates');

# Define some shorthand notation for ease of reference.
aLeft          = maze_action_enum.aLeft
aUp            = maze_action_enum.aUp
aRight         = maze_action_enum.aRight
aDown          = maze_action_enum.aDown

oNull          = maze_observation_enum.oNull
oLeftWall      = maze_observation_enum.oLeftWall
oUpWall        = maze_observation_enum.oUpWall
oRightWall     = maze_observation_enum.oRightWall
oDownWall      = maze_observation_enum.oDownWall

cUninformative = maze_observation_encoding_enum.cUninformative
cWalls         = maze_observation_encoding_enum.cWalls
cCoordinates   = maze_observation_encoding_enum.cCoordinates

# Define some maze layout constants.
cWall         = '@'
cTeleportTo   = '*'
cTeleportFrom = '!'
cEmpty        = '&'


class Maze(environment.Environment):
    """ A two-dimensional maze environment.

        The agent is able to move through the maze in each of the four cardinal directions
        (`aLeft`, `aUp`, `aRight`, `aDown`). The user is able to specify (via a
        configuration file) the dimensions, layout, and rewards of the maze as well
        as the type of observations given to the agent. In particular, the maze
        is a certain number of rows high and columns wide, each cell in the maze is
        of a certain type and has a certain reward.

        The type of each cell determines what happens to the agent when it attempts to move
        into the cell:

        - Wall (`cWall`): represents an impassable cell, attempting to move into
          a wall will result in the agent remaining at its current position.

        - Empty (`cEmpty`): an empty cell through which the agent can freely
          pass.

        - Teleport from (`cTeleportFrom`): represents a cell which, upon entry,
          will randomly teleport the agent to another cell of type `cTeleportTo`.

        - Teleport to (`cTeleportTo`): A cell which can be teleported to.
          Otherwise, the cell acts identically to an empty cell. The maze MUST
          contain at least one of these cells.

        The reward for each cell describes the reward received by the agent when it
        attempts to move into that cell. For example, when the agent attempts to move
        into a wall square, it gets the reward from the wall cell despite the fact
        that it does not end its turn in that cell. Similarly, when the agent moves
        into a `cTeleportFrom` cell it gets the reward from that cell rather
        than from the cell it will be teleported to.

        Finally, the user can choose one of several observation encodings to give
        to the agent:

        - Uninformative (`cUninformative`): The observation is a single unchanging value.

        - Walls (`cWalls`): The observation encodes the presence of walls in
          adjacent squares. This is encoded as a sum of the flags: `oDownWall`,
          `oLeftWall`, `oRightWall`, `oUpWall`, each of which indicates the presence
          of a wall in the corresponding direction.

        - Coordinates (`cCoordinates`): The observation gives the coordinates
          of the cell occupied by the agent. This is encoded as row * num_cols + col.

        Domain characteristic:

        - environment: "maze"

        - maximum action: 3 (2 bits)

        - maximum observation:

          - Uninformative: 0 (1 bit)

          - Walls: 15 (4 bits)

          - Coordinates: self.num_rows * self.num_cols - 1

        - maximum reward: maximum value in maze rewards (`self.max_reward`)

        Configuration options:

        - maze-num-rows: the number of rows in the maze.

        - maze-num-cols: the number of columns in the maze.

        - maze-rewards#: comma-separated list of rewards for each square in row #. If
          the agent enters (or attempts to enter) a particular square it receives
          the corresponding reward. 1 <= # <= maze-num-rows.

        - maze-layout#: The layout of row # of the maze (1 <= # <= maze-num-rows).

          Contains maze-num-cols symbols as follows:
          - `cWall`: indicates the square cannot be entered (a.k.a. "a wall").
          - `cTeleportTo`: indicates the square can be entered and teleported to.
          - `cEmpty`: indicates the square can be entered but not teleported to.
          - `cTeleportFrom`: indicates the square can be entered, and that doing
            so will randomly teleport the agent to a `cTeleportTo` square before
            the next turn (the agent receives the reward before the teleportation
            occurs).

        - maze-observation-encoding: Specifies the type of observations the agent receives
          - uninformative: the agent receives the same observation each cycle.
          - walls: the agent receives an observation specifying whether there are
                   walls "@" above, below, left, or right of its current position.
          - coordinates: the observation specifies the coordinates of the agent in
                         the maze.
    """

    # Instance methods.

    def __init__(self, options = {}):
        """ Construct the Maze environment from the given options.

             - `options` is a dictionary of named options and their values.
        """

        # Set up the base environment.
        environment.Environment.__init__(self, options = options)

        # Configure the environment based on the options.
        self.configure(options)

        # Define the acceptable action values.
        self.valid_actions = maze_action_enum.keys()

        # Define the acceptable observation values.
        self.valid_observations = xrange(0, self.max_observation() + 1)

        # Define the acceptable reward values.
        self.valid_rewards = xrange(0, self.max_reward + 1)

        # Assign the location of the agent randomly.
        self.teleport_agent()

        # Set the initial reward and percept.
        self.reward = 0
        self.calculate_observation()
    # end def

    def calculate_observation(self):
        """ Determines the observation to give to the agent based on its current
            location (`self.row` and `self_col`) and the observation encoding
            (`self.observation_encoding`). Store the observation in `self.observation`.

            (Called `calculate_observation` in C++ version.)
        """

        if self.observation_encoding == cUninformative:
            # Uninformative observation: agent always receives same observation.
            self.observation = oNull
        elif self.observation_encoding == cWalls:
            # Agent observes adjacent walls.
            self.observation = 0
            if self.col == 0 or self.maze_layout[self.row][self.col - 1] == cWall:
                self.observation += oLeftWall
            # end if
            if self.row == 0 or self.maze_layout[self.row - 1][self.col] == cWall:
                self.observation += oUpWall
            # end if
            if self.col + 1 == self.num_cols or self.maze_layout[self.row][self.col + 1] == cWall:
                self.observation += oRightWall
            # end if
            if self.row + 1 == self.num_rows or self.maze_layout[self.row + 1][self.col] == cWall:
                self.observation += oDownWall
            # end if
        elif self.observation_encoding == cCoordinates:
            # Agent observes the coordinates of its current square.
            self.observation = self.row * self.num_cols + self.col
        # end if
    # end def

    def configure(self, options):
        """ Configures the maze based on the given configuration options.
            May exit if the configuration is not validly formatted.
        """

        # Get the dimensions of the maze from the options.
        self.num_rows = options.get("maze-num-rows", None)
        if self.num_rows is None:
            sys.stderr.write("ERROR: configuration does not contain a 'maze-num-rows' value. Exiting." + \
                             os.linesep)
            sys.exit(1)
        # end if
        self.num_rows = int(self.num_rows)

        self.num_cols = options.get("maze-num-cols", None)
        if self.num_cols is None:
            sys.stderr.write("ERROR: configuration does not contain a 'maze-num-cols' value. Exiting." + \
                             os.linesep)
            sys.exit(1)
        # end if
        self.num_cols = int(self.num_cols)

        assert(self.num_rows > 0)
        assert(self.num_cols > 0)

        # Get the type of observations to give.
        encoding = options.get("maze-observation-encoding", "uninformative")
        if encoding == "uninformative":
            self.observation_encoding = cUninformative
        elif encoding == "walls":
            self.observation_encoding = cWalls
        elif encoding == "coordinates":
            self.observation_encoding = cCoordinates
        else:
            # We've got an unknown observation.
            sys.stderr.write("ERROR: Unknown observation encoding: '%s'" % str(encoding) + os.linesep)
            sys.exit(1)
        # end if

        # Variable indicating whether there are any squares which can be teleported to.
        teleport_impossible = True

        # Used to record the minimum and maximum rewards in the maze.
        min_reward = float('inf')
        self.max_reward = float('-inf')

        # Allocate and parse maze, determine if teleportation is possible and the
        # range of rewards.
        self.maze_rewards = {}
        self.maze_layout = {}
        self.teleport_to_locations = []
        for r in xrange(0, self.num_rows):
            # Get the reward string for the current row from the options.
            reward_option_name = "maze-rewards%d" % (r + 1)
            rewards = options.get(reward_option_name, None)

            if rewards is None:
                sys.stderr.write("ERROR: configuration does not contain a " + \
                                 "'%s' value as a num_rows value of '%d' implies. Exiting." % \
                                 (reward_option_name, self.num_rows) + os.linesep)
                sys.exit(1)
            # end if

            # Get the layout string for the current row from the options.
            layout_option_name = "maze-layout%d" % (r + 1)
            layout = options.get(layout_option_name, None)

            if layout is None:
                sys.stderr.write("ERROR: configuration does not contain a " + \
                                 "'%s' value as a num_rows value of '%d' implies. Exiting." % \
                                 (layout_option_name, self.num_rows) + os.linesep)
                sys.exit(1)
            # end if

            # Parse the reward and layout strings, and record appropriate properties of the maze.

            # Ensure the maze layout and rewards have a dictionary in each row to hold the
            # column values.
            if r not in self.maze_layout:
                self.maze_layout[r] = {}
            # end if

            if r not in self.maze_rewards:
                self.maze_rewards[r] = {}
            # end if

            # Split the layout string into a list, and split the reward strings into a list, splitting on commas.
            layout_list = list(layout)
            rewards_list = rewards.split(',')

            # If we don't have exactly enough entries, report this to the user.
            if len(layout_list) != self.num_cols:
                sys.stderr.write("ERROR: configuration value '%s' (%s)" % (layout_option_name, layout) + \
                                 "contains too %s entries. (Needs '%d'.) Exiting." % \
                                 ("few" if len(layout_list) < self.num_cols else "many", self.num_cols) + \
                                 os.linesep)
                sys.exit(1)
            # end if

            if len(rewards_list) != self.num_cols:
                sys.stderr.write("ERROR: configuration value '%s' (%s)" % (reward_option_name, rewards) + \
                                 "contains too %s entries. (Needs '%d'.) Exiting." % \
                                 ("few" if len(rewards_list) < self.num_cols else "many", self.num_cols) + \
                                 os.linesep)
                sys.exit(1)
            # end if

            # Turn the layout and reward strings into dictionary/two-dimensional array entries,
            # checking each value as it's inspected.
            for c in xrange(0, self.num_cols):
                #
                this_layout = layout_list[c]
                this_reward = int(rewards_list[c])

                self.maze_layout[r][c]  = this_layout
                self.maze_rewards[r][c] = this_reward

                # Is this the teleport-to character?
                if this_layout == cTeleportTo:
                    # Yes. Teleporting is possible in this maze.
                    teleport_impossible = False

                    # Record this location in the teleport-to location list.
                    self.teleport_to_locations += [(r, c)]
                # end if

                # Update the minimum and maximum reward if required.
                min_reward = min_reward if min_reward < this_reward else this_reward
                self.max_reward = self.max_reward if self.max_reward > this_reward else this_reward
            # end for
        # end for

        # Exit with an error message if it is impossible for the agent to teleport anywhere.
        if teleport_impossible:
            sys.stderr.write("ERROR: There must be at least one square the agent can teleport to.")
            sys.exit(1)
        # end if

        # Adjust rewards so they begin at 0.
        self.max_reward -= min_reward
        for r in xrange(0,  self.num_rows):
            for c in xrange(0, self.num_cols):
                self.maze_rewards[r][c] = self.maze_rewards[r][c] - min_reward
            # end for
        # end for
    # end def

    def max_observation(self):
        """ Returns the maximum observation that can be given to the agent.
            Depends on the observation encoding (`self.observation_encoding`) and
            (potentially) the dimensions of the maze (`self.num_rows` and `self.num_cols`.)

            (Called `max_observation` in the C++ version.)
        """
        if self.observation_encoding == cUninformative:
            # Only one observation
            return oNull
        elif self.observation_encoding == cWalls:
            # Maximum observation is walls on all sides
            return oLeftWall + oUpWall + oRightWall + oDownWall
        elif self.observation_encoding == cCoordinates:
            # Maximum observation is square at intersection of last row and column
            return (self.num_rows * self.num_cols) - 1
        # end if
    # end def

    def perform_action(self, action):
        """ Receives the agent's action and calculates the new environment percept.

            (Called `performAction` in the C++ version.)
        """

        assert self.is_valid_action(action)

        # Save the action.
        self.action = action

        # Set the flags for this action.
        self.teleported = False
        self.wall_collision = False

        # Calculate the square the agent is attempting to move to, making sure they
        # don't move outside the maze.
        self.row_to = (-1 if action == aUp else 0) + (1 if action == aDown else 0)
        self.row_to = min(max(self.row_to + self.row, 0), self.num_rows - 1)
        self.col_to = (-1 if action == aLeft else 0) + (1 if action == aRight else 0)
        self.col_to = min(max(self.col_to + self.col, 0), self.num_cols - 1)

        # Move the agent, making sure they don't walk into a wall.
        self.wall_collision = (self.maze_layout[self.row_to][self.col_to] == cWall)
        if not self.wall_collision:
            self.row = self.row_to
            self.col = self.col_to
        # end if

        # Teleport if appropriate.
        if self.maze_layout[self.row][self.col] == cTeleportFrom:
            self.teleport_agent()
        # end if

        # Calculate the reward for the square the agent *attempted* to move into,
        # regardless of whether they were able to move into it.
        self.reward = self.maze_rewards[self.row_to][self.col_to]

        # Calculate the observation for the square the agent is now in.
        # That is, after any movement or teleportation has occurred.
        self.calculate_observation()

        return (self.observation, self.reward)
    # end def

    def print(self):
        """ Returns a string indicating the current state of the environment, including
            the current location, observation, reward, and maze layout.
        """

        message = "row = %d" % self.row + ", col = %d" % self.col + \
                  ", observation = %d" % self.observation + \
                  ", reward = %d" % self.reward + \
                  (", teleported" if self.teleported else "") + \
                  (", wall collision" if self.wall_collision else "") + os.linesep

        for r in xrange(0, self.num_rows):
            for c in xrange(0, self.num_cols):
                if self.row == r and self.col == c:
                    message += "A"
                else:
                    message += self.maze_layout[r][c]
                # end if
            # end for
            message += os.linesep
        # end for

        return message
    # end def

    def teleport_agent(self):
        """ Randomly places the agent at any `cTeleportTo` location.

           (Called `teleportAgent` in the C++ version.)
        """

        self.teleported = True

        # This is altered from the C++ version to be far more efficient.
        # (e.g. instead of random search of the maze, use a random choice
        #  over a pre-computed list of possible destinations.)
        self.row, self.col = random.choice(self.teleport_to_locations)
    # end def
# end class


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Defines an environment for a biased coin flip.
"""

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import random

from pyaixi import environment, util

# Define a enumeration to represent coin flip actions, which is a prediction of the coin landing on
# heads or tails.
coin_flip_action_enum = util.enum('aTails', 'aHeads')

# Define a enumeration to represent coin flip observations e.g. the coin landed on heads or tails.
coin_flip_observation_enum = util.enum('oTails', 'oHeads')

# Define a enumeration to represent coin flip rewards e.g. win or lose, for correcting predicting
# the coin flip.
coin_flip_reward_enum = util.enum('rLose', 'rWin')

# Define some shorthand notation for ease of reference.
aHeads = coin_flip_action_enum.aHeads
aTails = coin_flip_action_enum.aTails

oHeads = coin_flip_observation_enum.oHeads
oTails = coin_flip_observation_enum.oTails

rLose = coin_flip_reward_enum.rLose
rWin = coin_flip_reward_enum.rWin

class CoinFlip(environment.Environment):
    """ A biased coin is flipped and the agent is tasked with predicting how it
        will land. The agent receives a reward of `rWin` for a correct
        prediction and `rLoss` for an incorrect prediction. The observation
        specifies which side the coin landed on (`oTails` or `oHeads`).
        The action corresponds to the agent's prediction for the
        next coin flip (`aTails` or `aHeads`).

        Domain characteristics:

        - environment: "coin_flip"
        - maximum action: 1 (1 bit)
        - maximum observation: 1 (1 bit)
        - maximum reward: 1 (1 bit)

        Configuration options:
        - `coin-flip-p`: the probability the coin lands on heads
                         (`oHeads`). Must be a number between 0 and 1 inclusive.
                         Default value is `default_probability`.
                         (Optional.)
    """

    # Instance attributes.

    # Set the default probability for the biased coin, when none is supplied from the options.
    default_probability = 0.7

    # Instance methods.

    def __init__(self, options = {}):
        """ Construct the CoinFlip environment from the given options.

             - `options` is a dictionary of named options and their values.

            The following options in `options` are optional:
             - `coin-flip-p`: the probability that the coin will land on heads. (Defaults to 0.7.)
        """

        # Set up the base environment.
        environment.Environment.__init__(self, options = options)

        # Define the acceptable action values.
        self.valid_actions = coin_flip_action_enum.keys()

        # Define the acceptable observation values.
        self.valid_observations = coin_flip_observation_enum.keys()

        # Define the acceptable reward values.
        self.valid_rewards = coin_flip_reward_enum.keys()

        # Determine the probability of the coin landing on heads.
        if 'coin-flip-p' not in options:
            options["coin-flip-p"] = self.default_probability
        # end if
        self.probability = float(options["coin-flip-p"])

        # Make sure the probability value is valid.
        assert 0.0 <= self.probability and self.probability <= 1.0

        # Set an initial percept.
        self.observation = oHeads if random.random() < self.probability else oTails
        self.reward = 0
    # end def

    def perform_action(self, action):
        """ Receives the agent's action and calculates the new environment percept.
            (Called `performAction` in the C++ version.)
        """

        assert self.is_valid_action(action)

        # Save the action.
        self.action = action

        # Flip the coin, set observation and reward appropriately.
        if (random.random() < self.probability):
            observation = oHeads
            reward = rWin if action == oHeads else rLose
        else:
            observation = oTails
            reward = rWin if action == oTails else rLose
        # end if

        # Store the observation and reward in the environment.
        self.observation = observation
        self.reward = reward

        return (observation, reward)
    # end def

    def print(self):
        """ Returns a string indicating the status of the environment.
        """

        message = "prediction: " + \
                  ("tails" if self.action == aTails else "heads") + \
                  ", observation: " + \
                  ("tails" if self.observation == oTails else "heads") + \
                  ", reward: %d" % self.reward

        return message
    # end def
# end class



import os

# Usage: python setup.py install
#        to install this package in your Python environment.
#
# Note for future packages updates:
#
# Use 'python setup.py check sdist --formats=gztar,zip upload' to update PyPi with new
# .tar.gz and .zip release files.

from distutils.core import setup

setup(
    name = 'pyaixi',
    version = '1.0.0',
    author = 'Geoff Kassel',
    author_email = 'geoffkassel_at_gmail_dot_com',
    package_dir = {'pyaixi': os.curdir},
    packages = ['pyaixi', 'pyaixi.agents', 'pyaixi.environments', 'pyaixi.prediction', 'pyaixi.search'],
    scripts = ['aixi.py'],
    url = 'https://github.com/gkassel/pyaixi',
    license = 'Creative Commons Attribution-ShareAlike 3.0 Unported License',
    description = 'A pure Python implementation of the Monte Carlo-AIXI-Context Tree Weighting (MC-AIXI-CTW) artificial intelligence algorithm.',
    long_description = open('README.md').read(),
    keywords = ["AIXI", "agent", "artificial intelligence", "general learning", "machine learning",
                "MC-AIXI-CTW", "model based", "reinforcement learning", "prediction", "search"],
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'License :: Freely Distributable',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Defines an environment for an agent playing Rock Paper Scissors against the environment.
"""

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import random

from pyaixi import environment, util

# Define a enumeration to represent rock-paper-scissors actions, which is the
# agent performing either a rock, paper, or a scissors move.
rock_paper_scissors_action_enum = util.enum('aRock', 'aPaper', 'aScissors')

# Define a enumeration to represent rock-paper-scissors observations, which is the
# opponent performing either a rock, paper, or a scissors move.
rock_paper_scissors_observation_enum = util.enum('oRock', 'oPaper', 'oScissors')

# Define a enumeration to represent losing, drawing, or winning.
rock_paper_scissors_reward_enum = util.enum('rLose', 'rDraw', 'rWin')

# Define some shorthand notation for ease of reference.
aRock     = rock_paper_scissors_action_enum.aRock
aPaper    = rock_paper_scissors_action_enum.aPaper
aScissors = rock_paper_scissors_action_enum.aScissors

oRock     = rock_paper_scissors_observation_enum.oRock
oPaper    = rock_paper_scissors_observation_enum.oPaper
oScissors = rock_paper_scissors_observation_enum.oScissors

rLose     = rock_paper_scissors_reward_enum.rLose
rDraw     = rock_paper_scissors_reward_enum.rDraw
rWin      = rock_paper_scissors_reward_enum.rWin


class RockPaperScissors(environment.Environment):
    """ The agent repeatedly plays Rock-Paper-Scissor against an opponent that has
        a slight, predictable bias in its strategy.

        If the opponent has won a round by playing rock on the previous cycle, it
        will always play rock at the next time step; otherwise it will pick an
        action uniformly at random.

        The agent's observation is the most recently chosen action of the opponent.
        It receives a reward of `rWin` for a win, `rDraw` for a draw and `rLose` for a loss.

        Domain characteristics:
         - environment: "rock_paper_scissors"
         - maximum action: 2 (2 bits)
         - maximum observation: 2 (2 bits)
         - maximum reward: 2 (2 bits)
    """

    # Instance methods.

    def __init__(self, options = {}):
        """ Construct the RockPaperScissors environment from the given options.

             - `options` is a dictionary of named options and their values.
        """

        # Set up the base environment.
        environment.Environment.__init__(self, options = options)

        # Define the acceptable action values.
        self.valid_actions = rock_paper_scissors_action_enum.keys()

        # Define the acceptable observation values.
        self.valid_observations = rock_paper_scissors_observation_enum.keys()

        # Define the acceptable reward values.
        self.valid_rewards = rock_paper_scissors_reward_enum.keys()

        # Set an initial percept.
        # (i.e. not rock, to ensure a random choice in the opponent on the first action.)
        self.observation = oPaper
        self.reward = 0
    # end def


    def perform_action(self, action):
        """ Receives the agent's action and calculates the new environment percept.
            (Called `performAction` in the C++ version.)
        """

        assert self.is_valid_action(action)

        # Save the action.  
        self.action = action

        # Opponent plays rock if it won the last round by playing rock, otherwise
        # it plays randomly.
        if self.observation == aRock and self.reward == rLose:
            self.observation = aRock
        else:
            self.observation = random.choice(self.valid_actions)
        # end if

        # Determine reward.
        if action == self.observation:
            # If both the agent and the opponent made the same move, it's a draw.
            self.reward = rDraw
        elif action == aRock:
            # If the opponent made a scissors move, then the agent wins if they played rock.
            self.reward = rWin if self.observation == oScissors else rLose
        elif action == aScissors:
            # If the opponent made a paper move, then the agent wins if they played scissors.
            self.reward = rWin if self.observation == oPaper else rLose
        elif action == aPaper:
            # If the opponent made a rock move, then the agent wins if they played paper.
            self.reward = rWin if self.observation == oRock else rLose
        # end if

        # Return the resulting observation and reward.
        return (self.observation, self.reward)
    # end def

    def print(self):
        """ Returns a string indicating the status of the environment.
        """

        action_text      = {aRock: "rock", aPaper: "paper", aScissors: "scissors"}
        observation_text = {oRock: "rock", oPaper: "paper", oScissors: "scissors"}
        reward_text      = {rLose: "loses", rDraw: "draws", rWin: "wins"}

        message = "Agent played " + action_text[self.action] + ", " + \
                  "environment played " + observation_text[self.observation] + "\t" + \
                  "Agent " + reward_text[self.reward]

        return message
    # end def
# end class